from django.test.utils import override_settings
from djangae.contrib.googleauth.backends.iap import IAPBackend
from djangae.contrib.googleauth.models import AnonymousUser
from djangae.test import TestCase
from unittest.mock import patch, Mock
from djangae.contrib.googleauth.middleware import AuthenticationMiddleware
from djangae.contrib.googleauth.backends.oauth2 import OAuthBackend
from djangae.contrib.googleauth.models import OAuthUserSession
from django.test import (RequestFactory)
from django.contrib.auth import get_user_model, BACKEND_SESSION_KEY


@override_settings(ROOT_URLCONF=__name__)
class AuthBackendTests(TestCase):

    def setUp(self):
        super().setUp()

        self.middleware = AuthenticationMiddleware()
        self.request = RequestFactory().get('/')
        self.request.session = {}
        self.mock_view = Mock(_auth_middleware_exempt=False)

        class ResolverMatch:
            func = self.mock_view

        self.request.resolver_match = ResolverMatch()

    @patch('djangae.contrib.googleauth.middleware.load_backend', return_value=OAuthBackend())
    def test_anonymous_user_if_not_authenticated(self, load_backend_mock):
        self.middleware.process_request(self.request)
        self.assertIsInstance(self.request.user, AnonymousUser)

    @patch('djangae.contrib.googleauth.middleware.IAPBackend.can_authenticate', return_value=True)
    @patch('djangae.contrib.googleauth.middleware.login')
    @patch('djangae.contrib.googleauth.middleware.get_backends')
    def test_user_if_iap_headers(self, get_backends_mock, login_mock, can_auth):
        iap_backend_mock = Mock(spec=IAPBackend)
        user = Mock()
        iap_backend_mock.authenticate.return_value = user
        get_backends_mock.return_value = [iap_backend_mock]

        self.middleware.process_request(self.request)

        can_auth.assert_called_once()
        iap_backend_mock.authenticate.assert_called_once()
        login_mock.assert_called_once_with(self.request, user)

    @patch('djangae.contrib.googleauth.middleware.IAPBackend.can_authenticate', return_value=True)
    @patch('djangae.contrib.googleauth.middleware.login')
    @patch('djangae.contrib.googleauth.middleware.get_backends')
    def test_no_iap_login_if_not_installed(self, get_backends_mock, login_mock, can_auth):
        oauth_backend_mock = Mock(spec=OAuthBackend)
        get_backends_mock.return_value = [oauth_backend_mock]

        self.middleware.process_request(self.request)

        can_auth.assert_not_called()
        login_mock.assert_not_called()

    @override_settings(GOOGLEAUTH_LINK_OAUTH_SESSION_EXPIRY=True)
    @patch('djangae.contrib.googleauth.middleware.get_user')
    @patch('djangae.contrib.googleauth.middleware.logout')
    @patch('djangae.contrib.googleauth.middleware.load_backend')
    def test_oauth_no_oauth_session(self, load_backend_mock, logout_mock, get_user_mock):
        # Session has an authenticated user
        user_mock = Mock(spec=get_user_model())
        user_mock.is_authenticated = True
        user_mock.google_oauth_id = '1'
        get_user_mock.return_value = user_mock

        # It has a session that uses oauth
        OAUTH_SESSION_KEY = 'key'
        self.request.session[BACKEND_SESSION_KEY] = OAUTH_SESSION_KEY

        # Loads the oauth backend
        oauth_backend_mock = Mock(spec=OAuthBackend)
        load_backend_mock.return_value = oauth_backend_mock

        self.middleware.process_request(self.request)

        # Check is loading the right backend
        load_backend_mock.assert_called_once_with(OAUTH_SESSION_KEY)

        # Session does not exist, logging out.
        logout_mock.assert_called_once_with(self.request)

    @override_settings(GOOGLEAUTH_LINK_OAUTH_SESSION_EXPIRY=True)
    @patch('djangae.contrib.googleauth.middleware.OAuthUserSession', spec=OAuthUserSession)
    @patch('djangae.contrib.googleauth.middleware.get_user')
    @patch('djangae.contrib.googleauth.middleware.logout')
    @patch('djangae.contrib.googleauth.middleware.load_backend')
    @patch('djangae.contrib.googleauth.middleware.redirect')
    @patch('djangae.contrib.googleauth.middleware.reverse', return_value="/login/")
    def test_oauth_oauth_session_invalid(self, reverse_mock, redirect_mock, load_backend_mock, logout_mock, get_user_mock, OAuthUserSession_mock): # noqa E501
        # Session has an authenticated user
        user_mock = Mock(spec=get_user_model())
        user_mock.is_authenticated = True
        user_mock.google_oauth_id = '1'
        get_user_mock.return_value = user_mock

        # It has a session that uses oauth
        OAUTH_SESSION_KEY = 'key'
        self.request.session[BACKEND_SESSION_KEY] = OAUTH_SESSION_KEY

        # Loads the oauth backend
        oauth_backend_mock = Mock(spec=OAuthBackend)
        load_backend_mock.return_value = oauth_backend_mock

        # Loads a OAuthUserSession that is invalid
        invalid_OAuth_session = Mock(OAuthUserSession)
        invalid_OAuth_session.is_valid = False
        OAuthUserSession_mock.objects.filter.return_value.first.return_value = invalid_OAuth_session

        self.middleware.process_request(self.request)

        # Check is getting user
        get_user_mock.assert_called_once_with(self.request)

        # Check is loading the right backend
        load_backend_mock.assert_called_once_with(OAUTH_SESSION_KEY)

        # Check is fetching the right session with
        OAuthUserSession_mock.objects.filter.assert_called_once_with(pk=self.request.user.google_oauth_id)

        # Session does not exist, logging out.
        logout_mock.assert_not_called()

        # Test redirects to login
        redirect_mock.assert_called_once_with("/login/?next=%2F")

    @override_settings(GOOGLEAUTH_LINK_OAUTH_SESSION_EXPIRY=True)
    @patch('djangae.contrib.googleauth.middleware.OAuthUserSession', spec=OAuthUserSession)
    @patch('djangae.contrib.googleauth.middleware.get_user')
    @patch('djangae.contrib.googleauth.middleware.logout')
    @patch('djangae.contrib.googleauth.middleware.load_backend')
    def test_oauth_oauth_session_valid(self, load_backend_mock, logout_mock, get_user_mock, OAuthUserSession_mock):
        # Session has an authenticated user
        user_mock = Mock(spec=get_user_model())
        user_mock.is_authenticated = True
        user_mock.google_oauth_id = '1'
        get_user_mock.return_value = user_mock

        # It has a session that uses oauth
        OAUTH_SESSION_KEY = 'key'
        self.request.session[BACKEND_SESSION_KEY] = OAUTH_SESSION_KEY

        # Loads the oauth backend
        oauth_backend_mock = Mock(spec=OAuthBackend)
        load_backend_mock.return_value = oauth_backend_mock

        # Loads a OAuthUserSession that is valid
        valid_OAuth_session = Mock(OAuthUserSession)
        valid_OAuth_session.is_valid = True
        OAuthUserSession_mock.objects.filter.return_value.first.return_value = valid_OAuth_session

        self.middleware.process_request(self.request)

        # Check is getting user
        get_user_mock.assert_called_once_with(self.request)

        # Check is loading the right backend
        load_backend_mock.assert_called_once_with(OAUTH_SESSION_KEY)

        # Check is fetching the right session with
        OAuthUserSession_mock.objects.filter.assert_called_once_with(pk=self.request.user.google_oauth_id)

        # Session does not exist, logging out.
        logout_mock.assert_not_called()

    @patch('djangae.contrib.googleauth.middleware.IAPBackend.can_authenticate', return_value=False)
    @patch('djangae.contrib.googleauth.middleware.get_user')
    @patch('djangae.contrib.googleauth.middleware.logout')
    @patch('djangae.contrib.googleauth.middleware.load_backend')
    def test_iap_iap_cant_authenticate(
        self,
        load_backend_mock,
        logout_mock,
        get_user_mock,
        can_authenticate_mock
    ):
        # Session has an authenticated user
        user_mock = Mock(spec=get_user_model())
        user_mock.is_authenticated = True
        get_user_mock.return_value = user_mock

        # It has a session_str
        OAUTH_SESSION_KEY = 'key'
        self.request.session[BACKEND_SESSION_KEY] = OAUTH_SESSION_KEY

        # Loads the iap backend
        iap_backend_mock = Mock(spec=IAPBackend)
        load_backend_mock.return_value = iap_backend_mock

        self.middleware.process_request(self.request)

        # Logging out.
        logout_mock.assert_called_once_with(self.request)

    @patch('djangae.contrib.googleauth.middleware.IAPBackend.can_authenticate', return_value=True)
    @patch('djangae.contrib.googleauth.middleware.get_user')
    @patch('djangae.contrib.googleauth.middleware.logout')
    @patch('djangae.contrib.googleauth.middleware.load_backend')
    def test_iap_iap_session_valid(
        self,
        load_backend_mock,
        logout_mock,
        get_user_mock,
        can_authenticate_mock
    ):
        # Session has an authenticated user
        user_mock = Mock(spec=get_user_model())
        user_mock.is_authenticated = True
        get_user_mock.return_value = user_mock

        # It has a session_str
        OAUTH_SESSION_KEY = 'key'
        self.request.session[BACKEND_SESSION_KEY] = OAUTH_SESSION_KEY

        # Loads the iap backend
        iap_backend_mock = Mock(spec=IAPBackend)
        load_backend_mock.return_value = iap_backend_mock

        self.middleware.process_request(self.request)

        # Logging out.
        logout_mock.not_called()



import threading
from typing import Optional

from django.core.exceptions import ImproperlyConfigured

_TASK_ENV = threading.local()


def _check_task_environment_middleware():
    from django.conf import settings  # nested to prevent early-import

    check1 = "djangae.tasks.middleware.task_environment_middleware"
    check2 = "djangae.tasks.middleware.TaskEnvironmentMiddleware"

    enabled = any([x in settings.MIDDLEWARE for x in (check1, check2)])

    if not enabled:
        raise ImproperlyConfigured(
            "You must add djangae.tasks.middleware.TaskEnvironmentMiddleware "
            "to your MIDDLEWARE setting to make use of djangae.tasks.environment"
        )


def is_in_task() -> bool:
    "Returns True if the request is a task, False otherwise"
    _check_task_environment_middleware()
    return bool(getattr(_TASK_ENV, "task_name", False))


def is_in_cron() -> bool:
    "Returns True if the request is in a cron, False otherwise"
    _check_task_environment_middleware()
    return bool(getattr(_TASK_ENV, "is_cron", False))


def task_name() -> Optional[str]:
    "Returns the name of the current task if any, else None"
    _check_task_environment_middleware()
    return getattr(_TASK_ENV, "task_name", None)


def task_retry_count() -> Optional[int]:
    "Returns the task retry count, or None if this isn't a task"
    _check_task_environment_middleware()
    try:
        return int(getattr(_TASK_ENV, "task_retry_count", None))
    except (TypeError, ValueError):
        return None


def task_queue_name() -> Optional[str]:
    "Returns the name of the current task queue (if this is a task) else 'default'"
    _check_task_environment_middleware()
    if is_in_task():
        return getattr(_TASK_ENV, "queue_name", "default")
    return None


def task_execution_count() -> Optional[int]:
    _check_task_environment_middleware()
    if is_in_task():
        return getattr(_TASK_ENV, "task_execution_count", 0)
    return None



import collections

from django.core.paginator import PageNotAnInteger, EmptyPage


class DatastorePaginator(object):
    """
    A paginator which only supports previous/next page controls and avoids doing
    expensive count() calls on datastore-backed queries.

    Does not implement the full Paginator API.
    """

    NOT_SUPPORTED_MSG = "Property '{}' is not supported when paginating datastore-models"

    def __init__(self, object_list, per_page, orphans=0,
                 allow_empty_first_page=True):
        self.object_list = object_list
        self.per_page = int(per_page)
        self.allow_empty_first_page = allow_empty_first_page

    def validate_number(self, number):
        """
        Validates the given 1-based page number.
        """
        try:
            number = int(number)
        except (TypeError, ValueError):
            raise PageNotAnInteger('That page number is not an integer')
        if number < 1:
            raise EmptyPage('That page number is less than 1')
        return number

    def page(self, number):
        """
        Returns a Page object for the given 1-based page number.
        """
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page

        fetched_objects = list(self.object_list[bottom:top + 1])
        has_next = len(fetched_objects) > self.per_page
        fetched_objects = fetched_objects[:self.per_page]

        return DatastorePage(has_next, fetched_objects, number, self)

    def _get_count(self):
        """
        Returns the total number of objects, across all pages.
        """
        raise NotImplementedError(self.NOT_SUPPORTED_MSG.format('count'))
    count = property(_get_count)

    def _get_num_pages(self):
        """
        Returns the total number of pages.
        """
        raise NotImplementedError(self.NOT_SUPPORTED_MSG.format('num_pages'))
    num_pages = property(_get_num_pages)

    def _get_page_range(self):
        """
        Returns a 1-based range of pages for iterating through within
        a template for loop.
        """
        raise NotImplementedError(self.NOT_SUPPORTED_MSG.format('page_range'))
    page_range = property(_get_page_range)


Paginator = DatastorePaginator


class DatastorePage(collections.Sequence):

    def __init__(self, has_next, object_list, number, paginator):
        self.object_list = object_list
        self._has_next = has_next
        self.number = number
        self.paginator = paginator

    def __repr__(self):
        bottom = (self.number - 1) * self.paginator.per_page
        top = len(self.object_list)
        return '<Objects {0} to {1}>'.format(bottom, top)

    def __len__(self):
        return len(self.object_list)

    def __getitem__(self, index):
        if not isinstance(index, (slice, int)):
            raise TypeError

        return self.object_list[index]

    def has_next(self):
        return self._has_next

    def has_previous(self):
        return self.number > 1

    def has_other_pages(self):
        return self.has_previous() or self.has_next()

    def next_page_number(self):
        return self.paginator.validate_number(self.number + 1)

    def previous_page_number(self):
        return self.paginator.validate_number(self.number - 1)

    def start_index(self):
        """
        Returns the 1-based index of the first object on this page,
        relative to total objects in the paginator.
        """
        # Special case, return zero if no items.
        if self.number == 1 and len(self.object_list) == 0:
            return 0
        return (self.paginator.per_page * (self.number - 1)) + 1

    def end_index(self):
        """
        Returns the 1-based index of the last object on this page,
        relative to total objects found (hits).
        """
        return (self.paginator.per_page * (self.number - 1)) + len(self.object_list)


Page = DatastorePage



import random
import string
from django.db import (
    connections,
    router,
)
from ..models import User


def _find_atomic_decorator(model):
    connection = connections[router.db_for_read(model)]

    # FIXME: When Django GCloud Connectors gets rid of its own atomic decorator
    # the Django atomic() decorator can be used regardless

    if connection.settings_dict['ENGINE'] == 'gcloudc.db.backends.datastore':
        from gcloudc.db.transaction import atomic
    else:
        from django.db.transaction import atomic

    return atomic


def _generate_unused_username(ideal):
    """
        Check the database for a user with the specified username
        and return either that ideal username, or an unused generated
        one using the ideal username as a base
    """

    if not User.objects.filter(username_lower=ideal.lower()).exists():
        return ideal

    exists = True

    # We use random digits rather than anything sequential to avoid any kind of
    # attack vector to get this loop stuck
    while exists:
        random_digits = "".join([random.choice(string.digits) for x in range(5)])
        username = "%s-%s" % (ideal, random_digits)
        exists = User.objects.filter(username_lower=username.lower).exists()

    return username



import logging
from datetime import datetime

from django.utils import dateparse

from . import indexers as search_indexers
from .tokens import tokenize_content


class IntegrityError(ValueError):
    pass


class Field(object):
    def __init__(self, default=None, null=True, index=True):
        """
            default: The default value for this field. A value of None or a string of
            non-indexable tokens will not be indexed, but the value will be stored on the document

            null: If False, will throw an IntegrityError if the value of the field is None. It will
                  *not* throw if the field contains a value that results in no indexable tokens, or
                  if the field is an empty string, only if the value is explicitly None.

            index: If False, the field will not be searchable, but will be stored on the document
        """
        self.default = default
        self.null = null
        self.index = index

    def normalize_value(self, value):
        # Default behaviour is to lower-case, remove punctuation
        # and then remove stop words

        if value is None:
            return None

        # Lower-case everything by default
        value = value.lower()

        # Normalize whitespace
        return " ".join(value.split())

    def tokenize_value(self, value):
        """
            Given a value set on a document, this
            returns a list of tokens that are indexed
        """
        if value is None:
            return value

        tokens, new_tokens = tokenize_content(value)
        tokens.extend(new_tokens)
        return tokens

    def clean_token(self, token):
        """
            Called on each token, if the token should be discarded,
            return None.
        """

        token = token.strip()  # Just in case

        # Remove + signs, unless they are trailing
        if "+" in token:
            plus_count = 0
            while len(token) > 0 and token[-1] == "+":
                token = token[:-1]
                plus_count += 1

            token = token.replace("+", "") + ("+" * plus_count)

        if "#" in token:
            # Replace hashes unless it's a music note or programming language
            if len(token) > 2 or token[-1] != "#" or token[0] not in "abcdefgjx":
                token = token.replace("#", "")

        return token

    def convert_from_index(self, value):
        """
            Convert a value returned from the index (these values)
            are stored in JSON format, so value will be a string, number, bool
            or None
        """
        return value


class AtomField(Field):
    pass


class TextField(Field):
    def normalize_value(self, value):
        if value is not None:
            value = str(value)

        return super().normalize_value(value)


class FuzzyTextField(TextField):
    DEFAULT_INDEXERS = (
        search_indexers.stemming,
    )

    def __init__(self, default=None, null=True, indexers=None, min_index_length=3, **kwargs):
        """
            indexers: list of indexers to apply to the value for indexing
            min_index_length: resulting tokens less than this length will be ignored
        """

        self.indexers = indexers or FuzzyTextField.DEFAULT_INDEXERS
        self.options = {
            "min_index_length": min_index_length
        }

        super().__init__(default=default, null=null, **kwargs)

    def tokenize_value(self, value):
        result = []
        tokens = super().tokenize_value(value)
        for token in tokens:
            for indexer in self.indexers:
                result.extend(indexer(token, **self.options))

        return result


class DateField(Field):
    def normalize_value(self, value):
        if value is None:
            return value

        if isinstance(value, str):
            return dateparse.parse_datetime(value)

        assert(isinstance(value, datetime))

        return value

    def tokenize_value(self, value):
        if value is None:
            return value

        # FIXME: It would be great to make this a datetime field
        # and handle times but that gets tricky quickly when you
        # consider timezones...
        return [
            value.strftime("%Y-%m-%d"),
        ]

    def convert_from_index(self, value):
        if value is None:
            return value

        try:
            return dateparse.parse_datetime(value)
        except ValueError:
            logging.warning("Unable to convert datetime back")
            return None


class NumberField(Field):
    def normalize_value(self, value):
        # FIXME: Validation?
        return int(value)

    def clean_token(self, value):
        return str(int(value))

    def tokenize_value(self, value):
        return [value]



import os

from django.db import models
from gcloudc.db import transaction

from djangae.contrib import sleuth
from djangae.tasks.deferred import defer
from djangae.test import (
    TaskFailedError,
    TestCase,
)


def test_task(*args, **kwargs):
    pass


def assert_cache_wiped(instance):
    field = DeferModelA._meta.get_field("b")
    assert(field.get_cached_value(instance, None) is None)


class DeferModelA(models.Model):
    b = models.ForeignKey("DeferModelB", on_delete=models.CASCADE)

    class Meta:
        app_label = "djangae"


class DeferModelB(models.Model):
    class Meta:
        app_label = "djangae"


class DeferModelC(models.Model):
    text = models.TextField()

    class Meta:
        app_label = "djangae"


def create_defer_model_b(key_value):
    DeferModelB.objects.create(pk=key_value)


def process_argument(arg):
    DeferModelC.objects.create(text=arg)


class DeferTests(TestCase):
    def test_large_task(self):
        random_file = os.path.join(os.path.dirname(__file__), "random_data")
        with open(random_file, "r") as f:
            big_string = f.read()

        try:
            defer(process_argument, big_string)
        except Exception:  # noqa
            self.fail("A large task couldn't be deferred")

        self.process_task_queues()

        self.assertTrue(DeferModelC.objects.exists())

        instance = DeferModelC.objects.get()
        self.assertEqual(instance.text, big_string)

    def test_wipe_related_caches(self):
        b = DeferModelB.objects.create()
        a = DeferModelA.objects.create(b=b)
        a.b  # Make sure we access it

        cache_name = DeferModelA._meta.get_field("b").get_cache_name()
        self.assertTrue(getattr(a, cache_name))

        defer(assert_cache_wiped, a)

        # Should raise an assertion error if the cache existed
        try:
            self.process_task_queues()
        except TaskFailedError as e:
            raise e.original_exception

        # Should not have wiped the cache for us!
        self.assertIsNotNone(getattr(a, cache_name, None))

    def test_queues_task(self):
        initial_count = self.get_task_count()
        defer(test_task)
        self.assertEqual(self.get_task_count(), initial_count + 1)

    def test_task_default_routing(self):
        gae_version = 'demo'
        os.environ['GAE_VERSION'] = gae_version

        with sleuth.watch('google.cloud.tasks_v2.CloudTasksClient.create_task') as _create_task:
            defer(test_task)

            self.assertTrue(_create_task.called)
            routing = _create_task.calls[0].args[2]['app_engine_http_request']['app_engine_routing']
            self.assertFalse('service' in routing)
            self.assertFalse('instance' in routing)
            self.assertEqual(routing['version'], gae_version)

        del os.environ['GAE_VERSION']

    def test_task_routing(self):
        service = 'service123'
        version = 'version456'
        instance = 'instance789'
        os.environ['GAE_VERSION'] = 'demo'

        with sleuth.watch('google.cloud.tasks_v2.CloudTasksClient.create_task') as _create_task:
            defer(test_task, _service=service, _version=version, _instance=instance)

            self.assertTrue(_create_task.called)
            routing = _create_task.calls[0].args[2]['app_engine_http_request']['app_engine_routing']
            self.assertEqual(routing['service'], service)
            self.assertEqual(routing['version'], version)
            self.assertEqual(routing['instance'], instance)

        del os.environ['GAE_VERSION']

    def test_deprecated_target_parameter(self):
        self.assertRaises(UserWarning, defer, test_task, _target='test')

    def test_transactional_defer(self):
        try:
            with transaction.atomic():
                defer(create_defer_model_b, 1, _transactional=True)
                raise ValueError()  # Rollback the transaction
        except ValueError:
            pass

        self.process_task_queues()

        # Shouldn't have created anything
        self.assertEqual(0, DeferModelB.objects.count())

        with transaction.atomic():
            defer(create_defer_model_b, 1, _transactional=True)
            self.process_task_queues()
            self.assertEqual(0, DeferModelB.objects.count())  # Still nothing

        # Now we should be good!
        self.process_task_queues()
        self.assertEqual(1, DeferModelB.objects.count())



from .fields import (
    Field,
    NumberField,
)


class Document(object):
    # All documents have an 'id' property, if this is blank
    # when indexing, it will be populated with a generated one
    # This corresponds with the PK of the underlying DocumentRecord
    id = NumberField()

    @property
    def pk(self):
        return self.id

    @property
    def revision(self):
        return self._record.revision if self._record else 0

    @property
    def persisted(self):
        return bool(self._record)

    @property
    def index_name(self):
        if not self.persisted:
            return None

        return self._record.index_stats_id

    def __init__(self, _record=None, **kwargs):
        self._record = _record

        if self._record:
            self.id = self._record.pk
        else:
            self.id = kwargs.get("id")

        self._fields = {}

        klass = type(self)

        for attr_name in dir(klass):
            attr = getattr(klass, attr_name)

            if isinstance(attr, Field):
                attr.attname = attr_name
                self._fields[attr_name] = attr

                # We set the ID value above based on _record or 'id'
                # and we don't want to wipe that
                if attr_name == "id":
                    continue

                # Apply any field values passed into the init
                if attr_name in kwargs:
                    setattr(self, attr_name, kwargs[attr_name])
                else:
                    # Set default if there was no value
                    setattr(self, attr_name, attr.default)

        for key in kwargs.keys():
            # Throw an error if the kwarg doesn't match a field
            if key not in self._fields:
                raise ValueError("Unknown field: %s" % key)

    def get_fields(self):
        return self._fields

    def get_field(self, name):
        return self._fields[name]

    def __eq__(self, other):
        return self.pk == other.pk

    def __repr__(self):
        return "<Document %s>" % self.pk



"""
    Tests for djangae.core.paginator to make sure that never does a full
    count on a query.
"""
from django.db import models

from djangae.contrib import sleuth
from djangae.test import TestCase
from djangae.core.paginator import Paginator


class SimplePaginatedModel(models.Model):
    field1 = models.IntegerField(default=0)
    field2 = models.CharField(max_length=10)

    class Meta:
        ordering = ("field1",)


class PaginatorTests(TestCase):
    def setUp(self):
        super(PaginatorTests, self).setUp()

        self.instances = [
            SimplePaginatedModel.objects.create(field1=x, field2=str(x+1))
            for x in range(10)
        ]

    def test_no_previous_on_first_page(self):
        with sleuth.watch('gcloudc.db.backends.datastore.commands.Query.__init__') as query:
            paginator = Paginator(SimplePaginatedModel.objects.all(), 2)

            self.assertFalse(query.called)
            page = paginator.page(1)
            self.assertFalse(page.has_previous())

            page = paginator.page(2)
            self.assertTrue(page.has_previous())

    def test_no_next_on_last_page(self):
        paginator = Paginator(SimplePaginatedModel.objects.all(), 2)
        page = paginator.page(5)
        self.assertTrue(page.has_previous())
        self.assertFalse(page.has_next())

        page = paginator.page(4)
        self.assertTrue(page.has_previous())
        self.assertTrue(page.has_next())

    def test_count_raises_error(self):
        paginator = Paginator(SimplePaginatedModel.objects.all(), 2)
        with self.assertRaises(NotImplementedError):
            paginator.count

    def test_num_pages_raises_error(self):
        paginator = Paginator(SimplePaginatedModel.objects.all(), 2)
        with self.assertRaises(NotImplementedError):
            paginator.num_pages

    def test_results_correct(self):
        paginator = Paginator(SimplePaginatedModel.objects.all(), 2)
        page = paginator.page(2)

        self.assertEqual(page[0].field1, 2)
        self.assertEqual(page[1].field1, 3)

        page = paginator.page(3)

        self.assertEqual(page[0].field1, 4)
        self.assertEqual(page[1].field1, 5)



# STANDARD LIB
from functools import wraps

# DJANGAE
from .kinds import LOCK_KINDS
from .memcache import MemcacheLock


class LockAcquisitionError(Exception):
    pass


class Lock(object):
    """ Common interface for acquiring and releasing a lock of the given kind.
        This is a lower-level interface than the `lock` decorator/context manager for cases where
        you need to acquire and release the lock(s) manually.
    """

    def __repr__(self):
        return u"<Lock (%s) '%s'>" % (self._kind, self._identifier)

    @classmethod
    def acquire(cls, identifier, wait=True, steal_after_ms=None, kind=LOCK_KINDS.STRONG, max_wait_ms=None):
        if kind == LOCK_KINDS.STRONG:
            from .models import DatastoreLock  # Avoid importing models before they're ready
            lock = DatastoreLock.objects.acquire(
                identifier, wait=wait, steal_after_ms=steal_after_ms, max_wait_ms=max_wait_ms
            )
        elif kind == LOCK_KINDS.WEAK:
            if max_wait_ms is not None:
                raise NotImplementedError("max_wait_ms is not implemented for WEAK locks")

            lock = MemcacheLock.acquire(
                identifier, wait=wait, steal_after_ms=steal_after_ms
            )
        else:
            raise Exception("Unsupported kind")

        if lock is None:
            return None

        instance = cls()
        instance._lock = lock
        # These two attributes are stored only for the benefit of the __repr__ method
        instance._identifier = identifier
        instance._kind = kind
        return instance

    def release(self):
        self._lock.release()


class LocknessMonster(object):
    """ Function decorator and context manager for locking a function or block of code so that only
        1 thread can excecute it at any given time.
        `identifier` should be a string which uniquely identifies the block of code that you want
        to lock.
        If `wait` is False and another thread is already running then:
            If used as a decorator, the function will not be run.
            If used as a context manager, LockAcquisitionError will be raised when entering `with`.
        If `steal_after_ms` is passed then existing locks on this function which are older
        than this value will be ignored.
    """

    def __init__(self, identifier, wait=True, steal_after_ms=None, kind=LOCK_KINDS.STRONG):
        self.identifier = identifier
        self.wait = wait
        self.steal_after_ms = steal_after_ms
        self.kind = kind

    def __call__(self, function):
        @wraps(function)
        def replacement_function(*args, **kwargs):
            try:
                with self:
                    return function(*args, **kwargs)
            except LockAcquisitionError:
                # In the case where self.wait is False and the Lock is already in use self.__enter__
                # will raise this exception
                return  # Do not run the function
        return replacement_function

    def __enter__(self):
        self.lock = Lock.acquire(self.identifier, self.wait, self.steal_after_ms, self.kind)
        if self.lock is None:
            raise LockAcquisitionError("Failed to acquire lock for '%s'" % self.identifier)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.lock:
            self.lock.release()
            self.lock = None  # Just for neatness


lock = LocknessMonster



import pickle

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from .decorators import task_only


@csrf_exempt
@task_only
def deferred_handler(request):
    callback, args, kwargs = pickle.loads(request.body)
    callback(*args, **kwargs)

    return HttpResponse("OK")



import copy
from functools import wraps

from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import Manager

from djangae.contrib import search
from djangae.contrib.search import fields as search_fields

from .index import Index


class ModelDocument(object):
    def __init__(self, model_class):
        self.model_class = model_class

    @classmethod
    def _meta(cls):
        class Meta(getattr(cls, "Meta")):
            @property
            def all_fields(self):
                additional_fields = []

                # Gather fields added directly on the model document as
                # overrides, rather than added on the fields list
                for attr_name in dir(cls):
                    attr = getattr(cls, attr_name, None)
                    if isinstance(attr, search_fields.Field):
                        additional_fields.append(attr_name)

                return list(self.fields) + [
                    x for x in additional_fields if x not in self.fields
                ]

        return Meta()

    @classmethod
    def index(cls):
        meta = cls._meta()
        index_name = getattr(meta, "index", "") if meta else None

        if not index_name:
            index_name = cls.__name__

        return Index(name=index_name)


def document_from_model_document(model, model_document):
    fields = model_document._meta().all_fields

    mapping = {
        models.AutoField: search_fields.NumberField,
        models.BigAutoField: search_fields.NumberField,
        models.BigIntegerField: search_fields.NumberField,
        models.BooleanField: search_fields.NumberField,
        models.CharField: search_fields.TextField,
        models.DateField: search_fields.DateField,
        models.DateTimeField: search_fields.DateField,
        models.DecimalField: search_fields.TextField,
        models.EmailField: search_fields.TextField,
        models.FloatField: search_fields.NumberField,
        models.IntegerField: search_fields.NumberField,
        models.PositiveIntegerField: search_fields.NumberField,
        models.PositiveIntegerField: search_fields.NumberField,
        models.PositiveSmallIntegerField: search_fields.NumberField,
        models.SlugField: search_fields.TextField,
        models.SmallIntegerField: search_fields.NumberField,
        models.TextField: search_fields.TextField,
        models.URLField: search_fields.TextField,
        models.UUIDField: search_fields.TextField,
    }

    pk_type = type(model._meta.pk)

    def get_search_field(model_field):
        try:
            return mapping[model_field]
        except KeyError:
            raise NotImplementedError("Unhandled model field type: %s", model_field)

    attrs = {
        "instance_id": get_search_field(pk_type)(index=False)
    }

    fields = list(fields) + ["instance_id"]

    for field in fields:
        # id is always stored as instance_id
        if field == "instance_id":
            continue

        try:
            field_type = type(model._meta.get_field(field))
        except FieldDoesNotExist:
            # This would happen if we added a field override for
            # a model field that didn't exist
            continue

        # First, do we have an override on the model document itself?
        if hasattr(model_document, field):
            # We copy, so we don't mess with the class version of this
            attrs[field] = copy.deepcopy(getattr(model_document, field))
        else:
            attrs[field] = get_search_field(field_type)()

    Document = type(
        '%sDocument' % model.__name__,
        (search.Document,),
        attrs
    )

    return Document


_registry = {}


class SearchManagerBase(object):
    pass


def register(model, model_document):
    default_manager = type(getattr(model, "objects", Manager()))

    if isinstance(default_manager, SearchManagerBase):
        # Already patched!
        # FIXME: Use the registry instead?
        return

    document_class = document_from_model_document(model, model_document)

    def _do_search(query, **options):
        """
            Return a list of model instance_ids from the results
            of the specified query
        """

        index = model_document.index()
        documents = index.search(query, document_class=document_class, **options)
        return [x.instance_id for x in documents]

    class SearchQueryset(models.QuerySet):
        def search(self, query, ordered_ids=None, **options):
            keys = _do_search(query, **options)
            if ordered_ids is not None:
                ordered_ids.extend(keys)
            return self.filter(pk__in=keys)

        def search_and_rank(self, query, **options):
            keys = _do_search(query, **options)

            return sorted(
                self.filter(pk__in=keys),
                key=lambda x: keys.index(x.pk)
            )

    class SearchManager(default_manager, SearchManagerBase):
        def get_queryset(self):
            qs = SearchQueryset(model, using=self._db)

            # Apply any filtering from any parent manager
            parent_qs = super().get_queryset()
            qs.query = parent_qs.query

            return qs

        def search(self, query, ordered_ids=None, **options):
            return self.get_queryset().search(
                query=query, ordered_ids=ordered_ids, **options
            )

        def search_and_rank(self, query, **options):
            return self.get_queryset().search_and_rank(
                query=query, **options
            )

    # FIXME: Is this safe? I feel like it should be but 'objects' is
    # a ManagerDescriptor so this might not be doing what I think it
    # is.
    model.objects.__class__ = SearchManager

    def delete_decorator(func):
        from djangae.contrib.search.models import DocumentRecord

        @wraps(func)
        def wrapped(self, *args, **kwargs):
            instance_id = self.pk

            func(self, *args, **kwargs)

            results = DocumentRecord.objects.filter(
                data__instance_id=instance_id
            ).values_list("pk", flat=True)

            for result in results:
                model_document.index().remove(result)

        return wrapped

    model.delete = delete_decorator(model.delete)

    def save_decorator(func):
        @wraps(func)
        def wrapped(self, *args, **kwargs):
            func(self, *args, **kwargs)

            # Force un-indexing before re-index
            delete_decorator(lambda self: None)(self)

            attrs = {
                f: model._meta.get_field(f).value_from_object(self)
                for f in model_document._meta().all_fields
            }

            attrs["instance_id"] = self.pk

            doc = document_class(**attrs)
            model_document.index().add(doc)
        return wrapped

    model.save = save_decorator(model.save)




from django.apps import AppConfig


class SearchConfig(AppConfig):
    name = 'djangae.contrib.search'
    verbose_name = "Search"

    def ready(self):
        super().ready()
        self.module.autodiscover()



from django.conf import settings
from djangae.contrib import sleuth
from djangae.test import TestCase
from djangae import environment
from ..utils import get_backup_setting, get_gcs_bucket


class GetDatastoreSettingTest(TestCase):

    def test_set_as_expected(self):
        with self.settings(DJANGAE_BACKUP_FOO=True):
            self.assertTrue(
                get_backup_setting('FOO')
            )

    def test_default(self):
        self.assertFalse(hasattr(settings, 'FOO'))
        self.assertEqual(
            get_backup_setting('FOO', required=False, default='bar'),
            'bar'
        )

    def test_not_required(self):
        """if a settings isnt required then return None"""
        self.assertFalse(hasattr(settings, 'FOO'))
        self.assertIsNone(
            get_backup_setting('FOO', required=False),
        )

    def test_required_when_settings_does_not_exist(self):
        with self.assertRaises(Exception):
            get_backup_setting('FOO')


class GetGcsBucketTest(TestCase):

    def test_custom_bucket_setting(self):
        with self.settings(DJANGAE_BACKUP_GCS_BUCKET='foo-bar-baz/qux'):
            result = get_gcs_bucket()

            self.assertEqual(result, 'foo-bar-baz/qux')

    def test_no_bucket_setting(self):
        with self.assertRaises(AttributeError):
            settings.DJANGAE_BACKUP_GCS_BUCKET

        result = get_gcs_bucket()

        self.assertEqual(result, 'example.appspot.com/djangae-backups')

    def test_no_bucket_setting_and_no_default_bucket(self):
        with self.assertRaises(AttributeError):
            settings.DJANGAE_BACKUP_GCS_BUCKET

        with sleuth.fake('djangae.environment.default_gcs_bucket_name', None):
            bucket = environment.default_gcs_bucket_name()

            self.assertIsNone(bucket)

            with self.assertRaisesRegexp(Exception, 'DJANGAE_BACKUP_GCS_BUCKET'):
                get_gcs_bucket()



import logging
import os

from django.apps import apps

import googleapiclient.discovery
from djangae.environment import (
    application_id,
    is_production_environment,
)
import google.auth
from google.oauth2 import service_account

from .utils import (
    get_backup_path,
    get_backup_setting,
)

logger = logging.getLogger(__name__)


AUTH_SCOPES = ['https://www.googleapis.com/auth/datastore']
SERVICE_URL = 'https://datastore.googleapis.com/$discovery/rest?version=v1'


def backup_datastore(bucket=None, kinds=None):
    """
    Using the new scheduled backup service write all required entity kinds
    to a specific GCS bucket path.
    """
    backup_enabled = get_backup_setting("ENABLED", False)
    if not backup_enabled:
        logger.warning(
            "DJANGAE_BACKUP_ENABLED is False or not set."
            "The datastore backup will not be run."
        )
        return

    # make sure no blacklisted entity kinds are included in our export
    valid_kinds = _get_valid_export_kinds(kinds)
    if not valid_kinds:
        logger.warning("No whitelisted entity kinds to export.")
        return

    # build the service object with the necessary credentials and trigger export
    service = _get_service()
    body = {
        'outputUrlPrefix': get_backup_path(bucket),
        'entityFilter': {
            'kinds': valid_kinds,
        }
    }
    app_id = application_id()
    request = service.projects().export(projectId=app_id, body=body)
    request.execute()


def _get_valid_export_kinds(kinds=None):
    """Make sure no blacklist models are included in our backup export."""
    excluded_models = get_backup_setting("EXCLUDE_MODELS", required=False, default=[])
    excluded_apps = get_backup_setting("EXCLUDE_APPS", required=False, default=[])

    to_backup = []
    for model in apps.get_models(include_auto_created=True):
        app_label = model._meta.app_label
        object_name = model._meta.object_name
        model_def = "{}_{}".format(app_label, object_name.lower())
        db_table = model._meta.db_table

        if app_label in excluded_apps:
            logger.info(
                "Not backing up %s due to the %s app being in DJANGAE_BACKUP_EXCLUDE_APPS",
                model_def, app_label
            )
            continue

        # Exclude the models if either the model label or datastore kind
        # is listed.
        if model_def in excluded_models or db_table in excluded_models:
            logger.info(
                "Not backing up %s as it is blacklisted in DJANGAE_BACKUP_EXCLUDE_MODELS",
                model_def
            )
            continue

        logger.info("%s added to list of models to backup", model_def)
        to_backup.append((model_def, db_table))

    # If kinds we explcitly provided by the caller, we only return those
    # already validated by our previous checks
    if kinds:
        to_backup = [
            kind for (_model_def, kind) in to_backup
            if _model_def in kinds or kind in kinds
        ]
    else:
        to_backup = [kind for (_model_def, kind) in to_backup]

    # If 2 models share the same underlying table they will return the same
    # kind. The datastore backup API will return a validation error when a kind
    # is listed more than once.
    return list(set(to_backup))


def _get_service():
    """Creates an Admin API service object for talking to the API."""
    credentials = _get_authentication_credentials()
    return googleapiclient.discovery.build(
        'admin', 'v1',
        credentials=credentials,
        discoveryServiceUrl=SERVICE_URL
    )


def _get_authentication_credentials():
    """
    Returns authentication credentials depending on environment. See
    https://developers.google.com/api-client-library/python/auth/service-accounts
    """
    if is_production_environment():
        credentials, _ = google.auth.default(scopes=AUTH_SCOPES)
    else:
        service_account_path = os.environ['GOOGLE_APPLICATION_CREDENTIALS']
        credentials = service_account.Credentials.from_service_account_file(
            service_account_path, scopes=AUTH_SCOPES
        )
    return credentials



import json

from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.shortcuts import reverse
from django.test import (
    LiveServerTestCase,
    RequestFactory,
    override_settings,
)
from django.urls import (
    include,
    path,
)
from django.utils import timezone

from djangae.contrib import sleuth
from djangae.contrib.googleauth.decorators import (
    login_required,
    oauth_scopes_required,
)
from djangae.contrib.googleauth.models import (
    AnonymousUser,
    OAuthUserSession,
    User,
)
from djangae.test import TestCase


class PermissionTests(TestCase):
    pass


@login_required(login_url='/googleauth/oauth2/login/')
def a_protected_view(request):
    return HttpResponse("OK")


urlpatterns = [
    path("protected", a_protected_view),
    path("googleauth/", include("djangae.contrib.googleauth.urls"))
]


class MockedAuth:
    def default(self):
        return MockedCreds(), 'test_project'


class MockedCreds:
    def __init__(self, *args, **kwargs):
        self._client_id = 'xxxxxxx'
        self._client_secret = 'yyyyyy'

    @property
    def client_id(self):
        return self._client_id

    @property
    def client_secret(self):
        return self._client_secret


@override_settings(ROOT_URLCONF=__name__)
class OAuthTests(LiveServerTestCase):

    def test_redirect_to_authorization_url(self):
        """
            Access to a page that is login_required
            should redirect to the authorization url
        """
        live_server_domain = self.live_server_url.split('://')[-1]
        protected_url = '/protected'
        response = self.client.get(protected_url, HTTP_HOST=live_server_domain)
        self.assertTrue(reverse("googleauth_oauth2login") in response.url)
        self.assertEqual(302, response.status_code)

        with sleuth.fake(
            'djangae.contrib.googleauth.views.OAuth2Session.authorization_url',
            return_value=('oauthauthurl', 'oauthstate')
        ) as auth_url:
            response = self.client.get(response.url, HTTP_HOST=live_server_domain)
            # check OAuthSession has been called properly
            self.assertTrue(auth_url.called)
            self.assertEqual(auth_url.calls[0].args[1], 'https://accounts.google.com/o/oauth2/v2/auth')

            self.assertTrue(set({
                "prompt": 'select_account',
                "include_granted_scopes": 'true',
            }.items()).issubset(set(auth_url.calls[0].kwargs.items())))

            self.assertEqual(set(json.loads(auth_url.calls[0].kwargs['state']).keys()), {'hostname', 'token'})

            # check session contains correct keys and values
            self.assertEqual(self.client.session.get('oauth-state'), 'oauthstate')
            self.assertEqual(self.client.session.get('next'), protected_url)

            # check that we're redirecting to authorization url returned from the session instance
            self.assertEqual(response.status_code, 302)
            self.assertTrue('oauthauthurl' in response.url)

    def test_oauth_callback_creates_session(self):
        """
            Should create an oauth session (if valid)
            and then redirect to the correct URL.

            Middleware should take care of authenticating the
            Django session
        """
        pass


@override_settings(ROOT_URLCONF=__name__)
class OAuth2CallbackTests(TestCase):

    def test_session_key_missing_raise_400(self):
        live_server_domain = self.live_server_url.split('://')[-1]
        response = self.client.get(reverse("googleauth_oauth2callback"), HTTP_HOST=live_server_domain)
        self.assertEqual(response.status_code, 400)

    def test_invalid_state_raises_400(self):
        live_server_domain = self.live_server_url.split('://')[-1]
        session = self.client.session
        session['oauth-state'] = 'somestate'
        session.save()
        response = self.client.get(reverse("googleauth_oauth2callback"), HTTP_HOST=live_server_domain)
        self.assertEqual(response.status_code, 400)

    def test_valid_credentials_log_user(self):
        live_server_domain = self.live_server_url.split('://')[-1]
        session = self.client.session
        session['oauth-state'] = 'somestate'
        session[REDIRECT_FIELD_NAME] = '/next_url'
        session.save()

        fake_token = {
            'access_token': '9999',
            'token_type': 'Bearer',
            'expires_in': '30',
            'scope': ['some.scope another.scope'],
            'id_token': 'someencryptedstuff'
        }

        fake_profile = {
            'id': '1',
            'email': 'test@example.com'
        }

        idinfo = {
            'sub': '1',
            'iss': 'accounts.google.com',
            'email': 'test@example.com'
        }

        with sleuth.fake('djangae.contrib.googleauth.views.OAuth2Session.fetch_token', return_value=fake_token), \
                sleuth.fake('djangae.contrib.googleauth.views.OAuth2Session.authorized', return_value=True), \
                sleuth.fake('djangae.contrib.googleauth.views.OAuth2Session.get', return_value=fake_profile), \
                sleuth.watch('djangae.contrib.googleauth.backends.oauth2.OAuthBackend.authenticate') as mocked_auth, \
                sleuth.watch('django.contrib.auth.login') as mocked_login, \
                sleuth.fake('google.oauth2.id_token.verify_token', idinfo):

            state = json.dumps({
                "hostname": "app.com",
            })
            response = self.client.get(
                reverse("googleauth_oauth2callback"),
                data={"state": state},
                HTTP_HOST=live_server_domain
            )

        # check authenticate and login function are called
        self.assertTrue(mocked_auth.called)
        self.assertTrue(mocked_login.called)

        self.assertEqual(response.status_code, 302)
        self.assertTrue(session[REDIRECT_FIELD_NAME] in response.url)

        user = User.objects.get(email="test@example.com")
        self.assertFalse(user.has_usable_password())

    @patch('django.contrib.auth.login', autospec=True)
    @patch('django.contrib.auth.authenticate', autospec=True)
    @patch('djangae.contrib.googleauth.views.OAuth2Session', autospec=True)
    def test_unauthorized_credentials_redirect_to_login(self, mocked_session, mocked_auth, mocked_login):
        # set authorized to return False, this should cause a redirect to login, and restart the flow
        mocked_session_instance = mocked_session.return_value
        mocked_session_instance.authorized = False

        live_server_domain = self.live_server_url.split('://')[-1]
        session = self.client.session
        session['oauth-state'] = 'somestate'
        session[REDIRECT_FIELD_NAME] = '/next_url'
        session.save()

        state = json.dumps({
            "hostname": "app.com",
        })
        response = self.client.get(
            reverse("googleauth_oauth2callback"),
            data={"state": state},
            HTTP_HOST=live_server_domain
        )

        # check authenticate and login function are not called
        self.assertFalse(mocked_auth.called)
        self.assertFalse(mocked_login.called)
        # check we're restarting login flow
        self.assertEqual(response.status_code, 302)
        self.assertTrue("/next_url" in response.url)


def a_view(request, *args, **kwargs):
    return HttpResponse(status=200)


@override_settings(ROOT_URLCONF=__name__)
class OAuthScopesRequiredTests(TestCase):
    def setUp(self):
        self._DEFAULT_OAUTH_SCOPES = [
                "https://www.googleapis.com/auth/userinfo.email",
                "https://www.googleapis.com/auth/userinfo.profile"
            ]
        self.factory = RequestFactory()
        self.oauthsession = OAuthUserSession.objects.create(
            id='1' * 21,
            scopes=self._DEFAULT_OAUTH_SCOPES,
            expires_at=timezone.now() + timedelta(seconds=10)
        )
        self.user = User.objects.create_user(
            google_oauth_id=self.oauthsession.pk,
            username='test',
            email='test@domain.com'
        )

    def test_oauth_scopes_required_call_view_if_no_additional_scopes(self):
        """When there are no additional scopes from the one in session, the view is simply called"""
        request = RequestFactory().get('/')
        request.user = self.user

        @oauth_scopes_required(scopes=[])
        def no_scopes_func(*args, **kwargs):
            no_scopes_func.called = True
            no_scopes_func.call_count += 1
        no_scopes_func.called = False
        no_scopes_func.call_count = 0

        @oauth_scopes_required(scopes=self._DEFAULT_OAUTH_SCOPES)
        def scopes_func(*args, **kwargs):
            scopes_func.called = True
            scopes_func.call_count += 1
        scopes_func.called = False
        scopes_func.call_count = 0

        no_scopes_func(request)

        self.assertTrue(no_scopes_func.called)
        self.assertEqual(no_scopes_func.call_count, 1)

        scopes_func(request)
        self.assertTrue(scopes_func.called)
        self.assertEqual(scopes_func.call_count, 1)

    def test_oauth_scopes_required_redirects_to_login_if_anonymous(self):
        request = RequestFactory().get('/')
        request.user = AnonymousUser()

        middleware = SessionMiddleware()
        middleware.process_request(request)
        request.session.save()

        def func(*args, **kwargs):
            func.called = True
        func.called = False

        decorated_func = oauth_scopes_required(scopes=[])(func)
        response_mocked = decorated_func(request)
        self.assertFalse(func.called)
        self.assertEquals(response_mocked.status_code, 302)
        self.assertTrue(reverse("googleauth_oauth2login") in response_mocked.url)

    def test_oauth_scopes_required_redirects_to_login_if_no_oauthsession(self):
        request = RequestFactory().get('/')
        request.user = User.objects.create_user(username='test2', email='test2@domain.com')

        middleware = SessionMiddleware()
        middleware.process_request(request)
        request.session.save()

        def func(*args, **kwargs):
            func.called = True
        func.called = False

        decorated_func = oauth_scopes_required(scopes=[])(func)
        response_mocked = decorated_func(request)
        self.assertFalse(func.called)
        self.assertEquals(response_mocked.status_code, 302)
        self.assertTrue(reverse("googleauth_oauth2login") in response_mocked.url)

    def test_oauth_scopes_required_redirects_for_additional_scopes(self):
        scopes = self._DEFAULT_OAUTH_SCOPES + ['https://www.googleapis.com/auth/calendar']
        request = RequestFactory().get('/')
        request.user = self.user

        middleware = SessionMiddleware()
        middleware.process_request(request)
        request.session.save()

        def func(*args, **kwargs):
            func.called = True
        func.called = False

        decorated_func_mock = oauth_scopes_required(scopes=scopes)(func)
        response_mocked = decorated_func_mock(request)
        self.assertFalse(func.called)
        # check we're redirecting to login url with the correct parameters
        self.assertEquals(response_mocked.status_code, 302)
        self.assertTrue(reverse("googleauth_oauth2login") in response_mocked.url)


class AuthBackendTests(TestCase):
    def test_valid_oauth_session_creates_django_session(self):
        pass

    def test_invalid_oauth_session_logs_out_django(self):
        pass

    def test_backend_does_nothing_if_authed_with_different_backend(self):
        """
            If you have the model backend and oauth backend (for example)
            then we don't log someone out if they authed with the model
            backend

        """
        pass



from .constants import (
    PUNCTUATION,
    WORD_DOCUMENT_JOIN_STRING,
)


def is_digit_or_single_char(token):
    """
    Acronyms or dates are formed either by tokens that have a sigle letter (acronyms ie. I-B-M)
    or numerical strings that can have more than 1 digit. (ie 2020-10-10).
    """
    return token.isdigit() or len(token) < 2


def tokenize_content(content):
    """
        We inherit the rules from the App Engine Search API
        when it comes to punctuation.

        We have a list of punctuation chars which break the
        content into tokens, and then some special cases where it
        makes sense. You can find the rules documented here:
        https://cloud.google.com/appengine/docs/standard/python/search#special-treatment
    """

    tokens = []
    current = ""

    STOP_CHARS = list(PUNCTUATION) + [" "]

    for c in content:
        if c in STOP_CHARS:
            if current.strip():
                tokens.append(current)

            if c.strip() and c != WORD_DOCUMENT_JOIN_STRING:
                tokens.append(c)

            current = ""
        else:
            current += c
    else:
        if current.strip():
            tokens.append(current)

    new_tokens = []
    tokens_to_append = []
    indexes_to_remove = []

    ACRONYM_TOKENS = (".", "-")
    current_at = None
    # Detect acronyms
    acronym_run = 0
    for i, token in enumerate(tokens):
        if (
            ((acronym_run and token == current_at) or (not acronym_run and token in ACRONYM_TOKENS)) and
                i > 0 and tokens[i - 1] != token and is_digit_or_single_char(tokens[i - 1])
        ):
            acronym_run += 1
            if acronym_run == 1:
                current_at = token
        else:
            if acronym_run > 1 and token != current_at:
                start = i - (2 * acronym_run)

                original = "".join(tokens[start:start + (acronym_run * 2) + 1])
                parts = [tokens[start + (x * 2)] for x in range(acronym_run + 1)]
                acronym = "".join(parts)

                # Add variations of the acronym
                new_tokens.append(acronym)
                new_tokens.append(".".join(parts))
                new_tokens.append("-".join(parts))

                # Remove the original characters
                indexes_to_remove.extend(range(start, start + (acronym_run * 2) + 1))

                if original in new_tokens:
                    new_tokens.remove(original)

                # Extend a single token made up of the whole original acronym
                # rather than seperate chars
                tokens_to_append.append(original)

                acronym_run = 0
            elif i > 0 and (tokens[i - 1] != current_at or not is_digit_or_single_char(token)):
                acronym_run = 0

    tokens = [x for i, x in enumerate(tokens) if i not in indexes_to_remove]
    tokens.extend(tokens_to_append)
    return tokens, new_tokens



import inspect
import re

from django.contrib.admindocs import utils as admindocs_utils
from django.contrib.admindocs import views as admindocs_views
from django.core.exceptions import ViewDoesNotExist
from django.urls.resolvers import URLResolver, URLPattern


def extract_views_from_urlpatterns(urlpatterns, base='', namespace=None, ignored_modules=None):
    """
    Return a list of views from a list of urlpatterns.

    Each object in the returned list is a tuple: (view_func, regex, name)
    """
    ignored_modules = ignored_modules if ignored_modules else []
    views = []

    for p in urlpatterns:

        if isinstance(p, URLPattern):
            # Handle correct single URL patterns
            try:
                if namespace:
                    name = '{0}:{1}'.format(namespace, p.name)
                else:
                    name = p.name
                if hasattr(p.callback, '__module__'):
                    if p.callback.__module__.split('.')[0] not in ignored_modules:
                        views.append((p.callback, base + str(p.pattern), name))
                else:
                    views.append((p.callback, base + str(p.pattern), name))
            except ViewDoesNotExist:
                continue

        elif isinstance(p, URLResolver):
            # Handle include() definitions
            try:
                patterns = p.url_patterns
            except ImportError:
                continue

            views.extend(
                extract_views_from_urlpatterns(
                    patterns, base + str(p.pattern),
                    namespace=(namespace or p.namespace),
                    ignored_modules=ignored_modules
                )
            )

        elif hasattr(p, '_get_callback'):
            # Handle string like 'foo.views.view_name' or just function view
            try:
                views.append((p._get_callback(), base + str(p.pattern), p.name))
            except ViewDoesNotExist:
                continue

        elif hasattr(p, 'url_patterns') or hasattr(p, '_get_url_patterns'):
            # Handle url_patterns objects
            try:
                patterns = p.url_patterns
            except ImportError:
                continue

            views.extend(
                extract_views_from_urlpatterns(
                    patterns, base + str(p.pattern),
                    namespace=namespace, ignored_modules=ignored_modules
                )
            )
        else:
            raise TypeError("%s does not appear to be a urlpattern object" % p)
    return views


def display_as_table(views, headers=('URL', 'Handler path', 'Decorators & Mixins')):
    """
        Get list of views from dumpurls security management command
        and returns them in the form of table to print in command line
    """
    # Find the longest value in each column
    widths = [len(max(columns, key=len)) for columns in zip(*[headers] + sorted(views))]
    widths = [width if width < 100 else 100 for width in widths]
    table_views = []

    table_views.append(
        ' | '.join('{0:<{1}}'.format(title, width) for width, title in zip(widths, headers))
    )
    table_views.append('-+-'.join('-' * width for width in widths))

    for row in views:
        for cell in row:
            if len(cell) > 100:
                cell = cell.split(',')
                cell = [",".join(cell[i:i+2]) for i in range(0, len(cell), 2)]

            mixins = cell
            if type(mixins) == list:
                for i, line in enumerate(mixins):
                    cell = line.strip()

        table_views.append(
            ' | '.join('{0:<{1}}'.format(cdata, width) for width, cdata in zip(widths, row))
        )

    return "\n".join([v for v in table_views]) + "\n"


def get_func_name(func):
    if hasattr(func, 'func_name'):
        return func.func_name
    elif hasattr(func, '__name__'):
        return func.__name__
    elif hasattr(func, '__class__'):
        return '%s()' % func.__class__.__name__
    else:
        return re.sub(r' at 0x[0-9a-f]+', '', repr(func))


def get_decorators(func):
    """
        Get function or class and return names of applied decorators.
        Note that due to the dynamic nature of python and the many ways in which functions can be
        decorated, patched or switched, this is - and will only ever be - a best effort attempt.
    """
    decorators = []
    if hasattr(func, '__module__'):
        mod = inspect.getmodule(func)
        source_code = inspect.getsourcelines(mod)[0]
        func_name = get_func_name(func)
        i = 0

        func_def = 'def {}'.format(func_name)
        class_def = 'class {}'.format(func_name)

        for line in source_code:
            if line.startswith(func_def) or line.startswith(class_def):
                j = 1
                k = source_code[i-j]
                # decorators can be defined on the previous line(s), but can have blank lines before them
                # blank lines are '\n', hence we strip() them to see if they're empty
                while k.startswith('@') or not k.strip():
                    if k.startswith('@'):
                        decorators.append(k.strip().split('(')[0])
                    j += 1
                    if j >= i:  # don't wrap around when we get to the start of the file
                        break
                    k = source_code[i-j]
            i += 1

    return decorators


def get_mixins(func, ignored_modules=None):
    """
        Given a Django class-based view, return names and paths to applied mixins
        Has an optional argument for names of modules that should be ignored.
        Note that there could be decorators on the methods of the class (or on methods of the
        mixins), but as with decorators, being sure of finding them all is impossible, so we simply
        list the mixins.
    """
    ignored_modules = ignored_modules if ignored_modules else []
    mixins = []
    # Django class-based views are used by calling .as_view() which creates a nested function that
    # has a 'cls' attribute which is the actual class
    if hasattr(func, 'cls'):
        for klass in func.cls.mro():
            if klass != func.cls and klass.__module__.split('.')[0] not in ignored_modules:
                mixins.append("{}.{}".format(klass.__module__, get_func_name(klass)))

    return mixins


# This is copied from django.contrib.admindocs.views, but improved to deal with non-capturing
# groups within the group.
# TODO: submit this as a patch to Django.

non_named_group_matcher = re.compile(
    r'\('  # opening bracket of group
    '('  # a group for THIS regex...
        r'[^\)]*'  # noqa zero or more characters that are not a closing bracket
        # the next line allows us to have non-capturing groups within the overall group that we're matching
        r'\(\?[^\)]*\)'  # a non-capturing group
        r'[^\)]*'  # zero or more characters that are not a bracket
    ')*'  # all of that ^ bit zero or more times
    r'\)'  # the closing bracket of the group
)


def simplify_regex(pattern):
    """ Do the same as django.contrib.admindocs.views.simplify_regex but with our improved regex.
    """

    original_regex = admindocs_utils.unnamed_group_matcher
    admindocs_utils.unnamed_group_matcher = non_named_group_matcher
    result = admindocs_views.simplify_regex(pattern)
    admindocs_utils.unnamed_group_matcher = original_regex

    return result



import logging

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from google.api_core import exceptions


class DjangaeTasksConfig(AppConfig):
    name = "djangae.tasks"
    verbose_name = "Djangae Tasks"

    def ready(self):
        """
            On startup we ensure the required queues
            exist based on settings.CLOUD_TASKS_QUEUES
        """
        from . import ensure_required_queues_exist
        try:
            ensure_required_queues_exist()
        except exceptions.ServiceUnavailable:
            logging.warning("Not creating required queues as Cloud Tasks client is unavailable")

        if not getattr(settings, "CLOUD_TASKS_LOCATION", None):
            raise ImproperlyConfigured(
                "You must specify settings.CLOUD_TASKS_LOCATION "
                "to use the djangae.tasks app."
            )



from functools import wraps


def additional_type_handler(func):
    @wraps(func)
    def _wrapper(self, o):
        if isinstance(o, set):
            # Return a string representing a set
            return "{" + ",".join([str(x) for x in o]) + "}"
        else:
            return func(self, o)

    return _wrapper


def patch():
    """
        This patches Djangos JSON encoder so it can deal with
        set(). This is necessary because otherwise we can't
        serialize SetFields
    """
    from django.core.serializers.json import DjangoJSONEncoder
    DjangoJSONEncoder.default = additional_type_handler(DjangoJSONEncoder.default)



from hashlib import md5
from django.conf import settings
from django.db import router
from django.core import paginator
from django.core.exceptions import FieldDoesNotExist
from django.core.cache import cache

from djangae.contrib.pagination.decorators import _field_name_for_ordering
from gcloudc.db.backends.datastore.query import extract_ordering


# TODO: it would be nice to be able to define a function which is given the queryset and returns
# the cache time.  That would allow different cache times for different queries.
CACHE_TIME = getattr(settings, "DJANGAE_PAGINATION_CACHE_TIME", 30 * 60)


class PaginationOrderingRequired(RuntimeError):
    pass


def _marker_cache_key(query_id, page_number):
    cache_key = "_PAGE_MARKER_{}:{}".format(query_id, page_number)
    return cache_key


def _count_cache_key(query_id):
    cache_key = "_PAGE_COUNTER_{}".format(query_id)
    return cache_key


def _update_known_count(query_id, count):
    cache_key = _count_cache_key(query_id)

    ret = cache.get(cache_key)
    if ret and ret > count:
        return

    cache.set(cache_key, count, CACHE_TIME)


def _get_known_count(query_id):
    cache_key = _count_cache_key(query_id)
    ret = cache.get(cache_key)
    if ret:
        return ret
    return 0


def _store_marker(query_id, page_number, marker_value):
    """
        For a model and query id, stores the marker value for previously
        queried page number.

        This stores the last item on the page identified by page number,
        not the marker that starts the page. i.e. there is a marker for page 1
    """

    cache_key = _marker_cache_key(query_id, page_number)
    cache.set(cache_key, marker_value, CACHE_TIME)


def _get_marker(query_id, page_number):
    """
        For a query_id, returns the marker at the end of the
        previous page. Returns a tuple of (marker, pages) where pages is
        the number of pages we had to go back to find the marker (this is the
        number of pages we need to skip in the result set)
    """

    counter = page_number - 1
    pages_skipped = 0

    while counter > 0:
        cache_key = _marker_cache_key(query_id, counter)
        ret = cache.get(cache_key)

        if ret:
            return ret, pages_skipped

        counter -= 1
        pages_skipped += 1

    # If we get here then we couldn't find a stored marker anywhere
    return None, pages_skipped


def queryset_identifier(queryset):
    """ Returns a string that uniquely identifies this query excluding its low and high mark"""

    db_alias = router.db_for_read(queryset.model)
    sql_and_params = u"%s:%s" % queryset.query.sql_with_params()
    cache_key = u"%s:%s" % (db_alias, sql_and_params)
    cache_key = cache_key.encode('utf-8')

    hasher = md5()
    hasher.update(cache_key)
    return hasher.hexdigest()


class Paginator(paginator.Paginator):
    """
        A paginator that works with the @paginated_model class decorator to efficiently
        return paginated sets on the appengine datastore
    """

    def __init__(
        self,
        object_list,
        per_page,
        readahead=10,
        allow_empty_first_page=True,
        **kwargs
    ):
        if not object_list.ordered:
            object_list.order_by("pk")  # Just order by PK by default

        self.original_orderings = extract_ordering(object_list.query)
        self.field_required = _field_name_for_ordering(self.original_orderings[:])
        self.readahead = readahead
        self.allow_empty_first_page = allow_empty_first_page

        try:
            object_list.model._meta.get_field(self.field_required)
        except FieldDoesNotExist:
            raise PaginationOrderingRequired(
                "No pagination ordering specified for {}. Field required: {}".format(
                    self.original_orderings,
                    self.field_required,
                )
            )

        # Wipe out the existing ordering
        object_list = object_list.order_by()

        # Add our replacement ordering
        # A single negated ordering can use the same field (we just flip the query), this
        # normalisation happens in _field_name_for_ordering, so we do the same here.
        if len(self.original_orderings) == 1 and self.original_orderings[0].startswith("-"):
            object_list = object_list.order_by("-" + self.field_required)
        else:
            object_list = object_list.order_by(self.field_required)

        self.queryset_id = queryset_identifier(object_list)
        super(Paginator, self).__init__(
            object_list,
            per_page,
            allow_empty_first_page=allow_empty_first_page,
            **kwargs
        )

    @property
    def count(self):
        return _get_known_count(self.queryset_id)

    def validate_number(self, number):
        """
        Validates the given 1-based page number.
        """
        try:
            number = int(number)
        except (TypeError, ValueError):
            raise paginator.PageNotAnInteger('That page number is not an integer')
        if number < 1:
            raise paginator.EmptyPage('That page number is less than 1')

        return number

    def page(self, number):
        """
        Returns a Page object for the given 1-based page number.
        """
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page

        marker_value, pages = _get_marker(
            self.queryset_id,
            number
        )

        if marker_value:
            if len(self.original_orderings) == 1 and self.original_orderings[0].startswith("-"):
                qs = self.object_list.all().filter(**{"{}__lt".format(self.field_required): marker_value})
            else:
                qs = self.object_list.all().filter(**{"{}__gt".format(self.field_required): marker_value})
            bottom = pages * self.per_page  # We have to skip the pages here
            top = bottom + self.per_page
        else:
            qs = self.object_list

        results = list(qs[bottom:top + (self.per_page * self.readahead)])

        next_page = results[top:]
        next_page_counter = number + 1
        while next_page:
            if len(next_page) > self.per_page-1:
                index = self.per_page-1
            else:
                index = len(next_page)-1
            _store_marker(
                self.queryset_id,
                next_page_counter,
                getattr(next_page[index], self.field_required)
            )
            next_page_counter += 1
            next_page = next_page[self.per_page:]

        if not results and not self.allow_empty_first_page:
            raise paginator.EmptyPage("That page contains no results")

        known_count = ((number - 1) * self.per_page) + len(results)
        _update_known_count(self.queryset_id, known_count)

        page = self._get_page(results[:self.per_page], number, self)

        if len(page.object_list) > self.per_page-1:
            index = self.per_page-1
        else:
            index = len(page.object_list)-1

        if results:
            _store_marker(
                self.queryset_id,
                number,
                getattr(page.object_list[index], self.field_required)
            )

        return page



from unittest.mock import patch
from djangae.tasks import deferred
from djangae.test import TestCase, TaskFailedBehaviour, TaskFailedError, CloudStorageTestCaseMixin


def my_task():
    """
    Basic task for testing task queues.
    """
    pass


def throw_once():
    throw_once.counter += 1
    if throw_once.counter == 1:
        raise Exception("First call")


throw_once.counter = 0


class TaskQueueTests(TestCase):

    def test_get_task_count(self):
        deferred.defer(my_task)
        deferred.defer(my_task, _queue='another')

        # We don't use self.assertNumTasksEquals here because we want to flush.
        task_count = self.get_task_count("default")
        self.assertEqual(1, task_count)

        task_count = self.get_task_count("another")
        self.assertEqual(1, task_count)

    def test_task_queue_processing_control(self):

        deferred.defer(throw_once)

        self.process_task_queues(failure_behaviour=TaskFailedBehaviour.RETRY_TASK)

        # Should've retried
        self.assertEqual(throw_once.counter, 2)

        throw_once.counter = 0

        deferred.defer(throw_once)

        self.assertRaises(
            TaskFailedError,
            self.process_task_queues,
            failure_behaviour=TaskFailedBehaviour.RAISE_ERROR
        )
        self.assertEqual(throw_once.counter, 1)


@patch('djangae.test.wipe_cloud_storage')
class CloudStorageTestCaseMixinTests(CloudStorageTestCaseMixin, TestCase):
    def test_wipes_on_setup(self, wipe_cloud_storage_mock):
        self.setUp()
        wipe_cloud_storage_mock.assert_called_once()



import os

from djangae.environment import (
    is_development_environment,
    is_production_environment,
)
from djangae.test import TestCase


class EnvironmentUtilsTest(TestCase):

    def test_is_production_environment(self):
        self.assertFalse(is_production_environment())
        os.environ["GAE_ENV"] = 'standard'
        self.assertTrue(is_production_environment())
        del os.environ["GAE_ENV"]

    def test_is_development_environment(self):
        self.assertTrue(is_development_environment())
        os.environ["GAE_ENV"] = 'standard'
        self.assertFalse(is_development_environment())
        del os.environ["GAE_ENV"]



"""
    This is duplicated from Django 3.0 to avoid
    starting an import chain that ends up with
    ContentTypes which may not be installed in a
    Djangae project.
"""


class BaseBackend:
    def authenticate(self, request, **kwargs):
        return None

    @classmethod
    def can_authenticate(cls, request):
        """
            This is a pre-check to see if the credentials are
            available to try to authenticate.
        """
        return True

    def get_user(self, user_id):
        return None

    def get_user_permissions(self, user_obj, obj=None):
        return set()

    def get_group_permissions(self, user_obj, obj=None):
        return set()

    def get_all_permissions(self, user_obj, obj=None):
        return {
            *self.get_user_permissions(user_obj, obj=obj),
            *self.get_group_permissions(user_obj, obj=obj),
        }

    def has_perm(self, user_obj, perm, obj=None):
        return perm in self.get_all_permissions(user_obj, obj=obj)



from django.urls import path

from . import views

urlpatterns = [
    path('start', views.start, name="instance_start"),
    path('stop', views.stop, name="instance_stop"),
    path('warmup', views.warmup, name="instance_warmup"),
    path('clearsessions', views.clearsessions, name="clearsessions"),
]



import random
from functools import partial

from gcloudc.db.models.fields.computed import ComputedCharField
from django.core.exceptions import FieldDoesNotExist

NULL_CHARACTER = u"\0"


def generator(fields, instance):
    """
        Calculates the value needed for a unique ordered representation of the fields
        we are paginating.
    """
    values = []
    for field in fields:
        neg = field.startswith("-")

        # If the field we have to paginate by is the pk, get the pk field name.
        if field == 'pk':
            field = instance._meta.pk.name

        value = instance._meta.get_field(field.lstrip("-")).value_from_object(instance)

        if hasattr(value, "isoformat"):
            value = value.isoformat()

        value = str(value)

        if neg:
            # this creates the alphabetical mirror of a string, e.g. ab => zy, but for the full
            # range of unicode characters, e.g. first unicode char => last unicode char, etc
            value = u"".join([chr(0xffff - ord(x)) for x in value])
        values.append(value)

    values.append(str(instance.pk) if instance.pk else str(random.randint(0, 1000000000)))

    return NULL_CHARACTER.join(values)


def _field_name_for_ordering(ordering):
    names = []

    # A single negated ordering can use the same field (we just flip the query)
    # so we normalize that out here and use the same field in that case
    if len(ordering) == 1 and ordering[0].startswith("-"):
        ordering = (ordering[0].lstrip("-"),)

    for field in ordering:
        if field.startswith("-"):
            names.append("neg_" + field[1:])
        else:
            names.append(field)

    new_field_name = "pagination_{}".format("_".join(names))
    return new_field_name


class PaginatedModel(object):
    """
        A class decorator which automatically generates pre-calculated fields for pagination.

        Specify the orderings (as a list of field tuples) you intend to paginate your models on, and
        additional fields will be added to the model which automatically calculate the appropriate
        value for pagination to work.

        The DatastorePaginator class will implicitly convert your query order_by to use these generated
        fields. Markers will be cached at the end of each page for each query, so that skipping pages
        is fast even when there are many pages.
    """
    def __init__(self, orderings):
        # Allow orderings to be specified either as single fields, or tuples/lists of fields
        _orderings = []
        for ordering in orderings:
            if isinstance(ordering, str):
                _orderings.append((ordering,))
            else:
                _orderings.append(ordering)
        self.orderings = _orderings

    def __call__(self, cls):
        """
            Dynamically adds pagination fields to a model depending on
            the orderings you specify
        """
        for ordering in self.orderings:
            new_field_name = _field_name_for_ordering(ordering)
            try:
                cls._meta.get_field(new_field_name)
            except FieldDoesNotExist:
                ComputedCharField(
                    partial(generator, ordering), max_length=500, editable=False
                ).contribute_to_class(cls, new_field_name)

        return cls


paginated_model = PaginatedModel



from django.apps import apps
from django.conf import settings
from django.db import models


CUSTOM_PERMISSIONS = getattr(settings, "GOOGLEAUTH_CUSTOM_PERMISSIONS", {})


def get_permission_choices():
    """
        Returns a list of permissions that can be set
        for users. Defaults are the same as Django
        (https://docs.djangoproject.com/en/3.0/topics/auth/default/#default-permissions)
    """

    DEFAULT_PERMISSIONS = ("add", "change", "delete", "view")
    GLOBAL_PERMISSIONS = tuple(
        list(DEFAULT_PERMISSIONS) + list(CUSTOM_PERMISSIONS.get('__all__', []))
    )

    result = []

    for app in apps.get_app_configs():
        for model in app.get_models():
            model_name = model.__name__
            app_model = "%s.%s" % (model._meta.app_label, model_name)

            codenames = list(GLOBAL_PERMISSIONS) + list(CUSTOM_PERMISSIONS.get(app_model, []))

            for permission in codenames:
                result.append(
                    (
                        "%s.%s_%s" % (
                            model._meta.app_label,
                            permission,
                            model_name.lower()
                        ),
                        "Can %s %s.%s" % (permission, model._meta.app_label, model_name)
                    )
                )

    return result


class PermissionChoiceIterator(object):
    def __iter__(self):
        for perm in get_permission_choices():
            yield perm


class PermissionChoiceField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 150
        kwargs["choices"] = PermissionChoiceIterator()
        # FIXME: for some reason, we're receiving two times verbose_name
        # and this is causing the method to fail
        if kwargs.get("verbose_name"):
            del kwargs["verbose_name"]
        super().__init__(self, *args, **kwargs)



from djangae.contrib import search
from djangae.contrib.search import fields
from djangae.contrib.search.model_document import document_from_model_document
from djangae.contrib.search.models import TokenFieldIndex
from djangae.test import TestCase

from .models import (
    SearchableModel1,
    SearchableModel2,
)


class SearchableModelDocument(search.ModelDocument):
    class Meta:
        index = "index1"
        fields = (
            "name",
        )

    other_thing = fields.NumberField()


class CharFieldPK(search.ModelDocument):
    class Meta:
        fields = (
            "sid",
        )


class SearchableTest(TestCase):
    def setUp(self):
        # Ensure that the model has been registered
        search.register(SearchableModel1, SearchableModelDocument)
        search.register(SearchableModel2, CharFieldPK)

        super().setUp()

        self.i1 = SearchableModel1.objects.create(id=1, name="Luke")
        self.i2 = SearchableModel1.objects.create(name="Jimmy")  # noqa
        self.i3 = SearchableModel1.objects.create(name="Paolo")  # noqa
        self.i4 = SearchableModel1.objects.create(name="Purvi")  # noqa
        self.i5 = SearchableModel1.objects.create(name="Alton Powers")
        self.instances = [self.i1, self.i2, self.i3, self.i4, self.i5]

    def test_pk_not_implicitly_searchable(self):
        """
            All ModelDocuments have an implicit instance_id. We need
            to make sure that isn't implicitly searchable
        """

        SearchableModel1.objects.create(id=999, name="test")
        results = list(SearchableModel1.objects.search("999"))
        self.assertFalse(results)

        results = list(SearchableModel1.objects.search("test"))
        self.assertTrue(results)

    def test_searching_models(self):
        results = SearchableModel1.objects.search("luke")
        self.assertCountEqual(results, [self.i1])

        results = SearchableModel1.objects.search("powers")
        self.assertCountEqual(results, [self.i5])

    def test_queryset_filtering(self):
        qs = SearchableModel1.objects.filter(
            pk__in=[self.i1.pk, self.i2.pk]
        ).search("luke")

        self.assertCountEqual(qs, [self.i1])

    def test_field_override(self):
        document = document_from_model_document(SearchableModel1, SearchableModelDocument)

        # The name field should be overridden, it would default to
        self.assertEqual(type(document.other_thing), fields.NumberField)

    def test_charfield_pk(self):
        i1 = SearchableModel2.objects.create(sid="test")
        i2 = SearchableModel2.objects.create(sid=1)  # Intentional integer
        i2.refresh_from_db()

        results = SearchableModel2.objects.search("test")
        self.assertCountEqual([i1], results)

        results = SearchableModel2.objects.search("sid:1")
        self.assertCountEqual([i2], results)

    def test_deletion(self):
        results = SearchableModel1.objects.search("jimmy")
        self.assertTrue([x for x in results])

        count = TokenFieldIndex.objects.count()
        idx = SearchableModelDocument.index()

        doc_count = idx.document_count()
        self.i2.delete()
        self.assertEqual(idx.document_count(), doc_count - 1)

        self.process_task_queues()

        new_count = TokenFieldIndex.objects.count()
        self.assertTrue(new_count < count)

        results = SearchableModel1.objects.search("jimmy")
        self.assertFalse([x for x in results])

    def test_update(self):
        results = SearchableModel1.objects.search("jimmy")
        self.assertTrue([x for x in results])

        self.i2.name = "bob"
        self.i2.save()

        results = SearchableModel1.objects.search("jimmy")
        self.assertFalse([x for x in results])

        results = SearchableModel1.objects.search("bob")
        self.assertTrue([x for x in results])

    def test_searching_stopwords(self):
        SearchableModel1.objects.create(name="About you")

        results = list(SearchableModel1.objects.search("about"))
        self.assertEqual(len(results), 1)

        results = list(SearchableModel1.objects.search("you"))
        self.assertEqual(len(results), 1)

    def test_search_ranking_applied(self):
        i1 = SearchableModel1.objects.create(name="testing")
        i2 = SearchableModel1.objects.create(name="test")
        i3 = SearchableModel1.objects.create(name="testy")

        ordered_ids = []

        results = list(
            SearchableModel1.objects.search(
                "test", use_startswith=True, ordered_ids=ordered_ids
            )
        )

        results = sorted(results, key=lambda x: ordered_ids.index(x.pk))

        self.assertEqual([i2, i3, i1], results)

        # Same as above
        results = SearchableModel1.objects.search_and_rank(
            "test", use_startswith=True
        )

        self.assertEqual([i2, i3, i1], results)

        # Same, but with the queryset instead of the manager
        results = SearchableModel1.objects.all().search_and_rank(
            "test", use_startswith=True
        )

        self.assertEqual([i2, i3, i1], results)




from django.test import LiveServerTestCase

from djangae.tasks import (
    cloud_tasks_parent_path,
    cloud_tasks_queue_path,
    ensure_required_queues_exist,
    get_cloud_tasks_client,
)
from google.api_core.exceptions import GoogleAPIError


class TaskFailedBehaviour:
    DO_NOTHING = 0
    RETRY_TASK = 1
    RAISE_ERROR = 2


class TaskFailedError(Exception):
    def __init__(self, task_name, status_code, original_exception=None):
        self.task_name = task_name
        self.status_code = status_code
        self.original_exception = original_exception

        super(TaskFailedError, self).__init__(
            "Task {} failed with status code: {}. \n\nMessage was: {}".format(
                task_name, status_code, original_exception
            )
        )


class TestCaseMixin(LiveServerTestCase):
    """
        A TestCase base class that manages task queues
        during testing. Ensures that required queues
        are created and paused, and manually runs the
        queued tasks in them to check their responses
    """
    def __init__(self, *args, **kwargs):
        self.max_task_retry_count = 100
        super().__init__(*args, **kwargs)

    def setUp(self):
        # Create all the queues required by this application

        super().setUp()

        # Find the port we were allocated
        self._server_port = self.live_server_url.rsplit(":")[-1]

        ensure_required_queues_exist()

        self.task_client = get_cloud_tasks_client()

        parent = cloud_tasks_parent_path()

        for queue in self.task_client.list_queues(parent=parent):
            # Make sure the queue is paused
            self.task_client.pause_queue(queue.name)

            # Make sure it's empty
            self.task_client.purge_queue(queue.name)

    def _get_queues(self, queue_name=None):
        if queue_name:
            path = cloud_tasks_queue_path(queue_name)
            queue = self.task_client.get_queue(path)
            queues = [queue]
        else:
            parent = cloud_tasks_parent_path()
            queues = self.task_client.list_queues(parent)

        return queues

    def flush_task_queues(self, queue_name=None):
        for queue in self._get_queues(queue_name=queue_name):
            self.task_client.purge_queue(queue.name)

    def get_task_count(self, queue_name=None):
        count = 0
        for queue in self._get_queues(queue_name=queue_name):
            path = queue.name
            count += len(list(self.task_client.list_tasks(path)))

        return count

    def assertNumTasksEquals(self, num, queue_name=None):
        self.assertEqual(num, self.get_task_count(queue_name=queue_name))

    def _get_all_tasks_for_queues(self, queue_names):
        tasks = []
        for path in queue_names:
            tasks += [x for x in self.task_client.list_tasks(path)]
        return tasks

    def process_task_queues(self, queue_name=None, failure_behaviour=TaskFailedBehaviour.RAISE_ERROR):
        queue_names = [q.name for q in self._get_queues(queue_name)]

        tasks = self._get_all_tasks_for_queues(queue_names)
        task_failure_counts = {}

        while tasks:
            task = tasks.pop(0)

            try:
                response = self.task_client.run_task(task.name + "?port=%s" % self._server_port)

                # If the returned status wasn't a success then
                # drop into the except block below to handle the
                # failure
                status = response.last_attempt.response_status.code
                if str(status)[0] != "2":
                    raise GoogleAPIError("Task returned bad status: %s" % status)

            except GoogleAPIError as e:
                if failure_behaviour == TaskFailedBehaviour.RETRY_TASK:
                    if task.name not in task_failure_counts:
                        task_failure_counts[task.name] = 1
                    else:
                        task_failure_counts[task.name] += 1

                    if task_failure_counts[task.name] >= self.max_task_retry_count:
                        # Make sure we don't get an infinite loop while retrying
                        raise

                    tasks.append(task)  # Add back to the end of the queue
                    continue
                elif failure_behaviour == TaskFailedBehaviour.RAISE_ERROR:
                    raise TaskFailedError(task.name, str(e))
                else:
                    # Do nothing, ignore the failure
                    pass

            if not tasks:
                tasks = self._get_all_tasks_for_queues(queue_names)



import uuid

from django.db import models

from gcloudc.db.models.fields.json import JSONField

from .document import Document
from .constants import WORD_DOCUMENT_JOIN_STRING


class DocumentRecord(models.Model):
    """
        'Document' is intentionally not a model;
        it would ruin the abstraction, and we need to
        store all kinds of data related to a Document.
        So instead, each Document has an instance of DocumentRecord
        this is the closest to a database representation of the doc
        and indeed, is where the document ID comes from.

        DocumentRecord exists to keep a reference to all token indexes
        and any stats/settings about the document (e.g. its rank).
    """
    index_stats = models.ForeignKey("IndexStats", on_delete=models.CASCADE)

    # The revision of the document. This gets changed on each reindex
    revision = models.UUIDField(default=uuid.uuid4, editable=False)

    # This is the data at the time the field was indexed so the doc
    # can be reconstructed on fetch
    data = JSONField()


class TokenFieldIndex(models.Model):
    # key should be of the format WWWW|XXXX|YYYY|ZZZZ where:
    # WWWW = index ID
    # XXXX = normalised token
    # YYYY = field_name
    # ZZZZ = document id

    # Querying for documents or fields containing the token
    # will just be a key__startswith query (effectively)
    id = models.CharField(primary_key=True, max_length=1500, default=None)

    index_stats = models.ForeignKey("IndexStats", on_delete=models.CASCADE)

    # We don't cascade delete when a document is deleted, because we clear up
    # in a background task
    record = models.ForeignKey("DocumentRecord", on_delete=models.DO_NOTHING)

    revision = models.UUIDField(default=None, editable=False, blank=False)

    token = models.CharField(max_length=500)
    field_name = models.CharField(max_length=500)

    @classmethod
    def generate_key(cls, index_name, token, field_name, document_id, revision):
        return WORD_DOCUMENT_JOIN_STRING.join(
            [str(x) for x in (index_name, token, field_name, document_id, revision)]
        )

    @classmethod
    def document_id_from_pk(cls, pk):
        """
            Given a PK in the right format, return the document ID
        """
        if pk is None:
            return None

        return int(pk.split(WORD_DOCUMENT_JOIN_STRING)[-2])

    @property
    def document_id(self):
        return self.record_id

    @property
    def document(self):
        return Document.objects.get(pk=self.document_id)

    def save(self, *args, **kwargs):
        assert(self.token.strip())  # Check we're not indexing whitespace or nothing
        assert(WORD_DOCUMENT_JOIN_STRING not in self.token)  # Don't index this special symbol

        orig_pk = self.pk

        self.pk = TokenFieldIndex.generate_key(
            self.index_stats_id, self.token, self.field_name, self.document_id, self.revision
        )
        # Just check that we didn't *change* the PK
        assert((orig_pk is None) or orig_pk == self.pk)
        super().save(*args, **kwargs)


class IndexStats(models.Model):
    """
        This is a representation of the index
        in the datastore. Its PK is used as
        a prefix to documents and token tables
        but it's only really used itself to maintain
        statistics about the indexed data.
    """

    name = models.SlugField(max_length=100, primary_key=True)
    document_count = models.PositiveIntegerField(default=0)



import random
import time
from datetime import datetime

from django.core.cache import cache


class MemcacheLock(object):
    def __init__(self, identifier, cache, unique_value):
        self.identifier = identifier
        self._cache = cache
        self.unique_value = unique_value

    @classmethod
    def acquire(cls, identifier, wait=True, steal_after_ms=None):
        start_time = datetime.utcnow()
        unique_value = random.randint(1, 100000)

        while True:
            acquired = cache.add(identifier, unique_value)
            if acquired:
                return cls(identifier, cache, unique_value)
            elif not wait:
                return None
            else:
                # We are waiting for the lock
                if (datetime.utcnow() - start_time).total_seconds() * 1000 > steal_after_ms:
                    # Steal anyway
                    cache.set(identifier, unique_value)
                    return cls(identifier, cache, unique_value)

                time.sleep(0)

    def release(self):
        cache = self._cache

        # Delete the key if it was ours. There is a race condition here
        # if something steals the lock between the if and the delete...
        if cache.get(self.identifier) == self.unique_value:
            cache.delete(self.identifier)



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation

from djangae.contrib.googleauth import (
    _GOOG_JWT_ASSERTION_HEADER,
    _GOOG_AUTHENTICATED_USER_EMAIL_HEADER,
    _GOOG_AUTHENTICATED_USER_ID_HEADER,
    _IAP_AUDIENCE
)
from djangae.contrib.googleauth.models import UserManager

from google.auth.transport import requests
from google.oauth2 import id_token
from . import _find_atomic_decorator, _generate_unused_username
from .base import BaseBackend

User = get_user_model()


class IAPBackend(BaseBackend):

    @classmethod
    def can_authenticate(cls, request):
        return _GOOG_AUTHENTICATED_USER_EMAIL_HEADER in request.META and \
            _GOOG_AUTHENTICATED_USER_EMAIL_HEADER in request.META and \
            _GOOG_JWT_ASSERTION_HEADER in request.META

    def authenticate(self, request, **kwargs):
        error_partial = 'An attacker might have tried to bypass IAP.'
        atomic = _find_atomic_decorator(User)
        user_id = request.META.get(_GOOG_AUTHENTICATED_USER_ID_HEADER)
        email = request.META.get(_GOOG_AUTHENTICATED_USER_EMAIL_HEADER)

        # User not logged in to IAP
        if not user_id or not email:
            return

        # All IDs provided should be namespaced
        if ":" not in user_id or ":" not in email:
            return

        # Google tokens are namespaced with "auth.google.com:"
        namespace, user_id = user_id.split(":", 1)
        _, email = email.split(":", 1)

        if not getattr(request, "_through_local_iap_middleware", False):
            try:
                audience = _get_IAP_audience_from_settings()
            except AttributeError:
                raise ImproperlyConfigured(
                    "You must specify a %s in settings when using IAPBackend" % (
                        _IAP_AUDIENCE,
                    ))
            iap_jwt = request.META.get(_GOOG_JWT_ASSERTION_HEADER)

            try:
                signed_user_id, signed_user_email = _validate_iap_jwt(iap_jwt, audience)
                signed_user_namespace, signed_user_id = signed_user_id.split(":", 1)
            except ValueError as e:
                raise SuspiciousOperation("**ERROR: JWT validation error {}**\n{}".format(e, error_partial))

            assert (signed_user_id == user_id), (
                    f"IAP signed user id does not match {_GOOG_AUTHENTICATED_USER_ID_HEADER}. ",
                    error_partial,
                )
            assert (signed_user_email == email), (
                    f"IAP signed user email does not match {_GOOG_AUTHENTICATED_USER_EMAIL_HEADER}. ",
                    error_partial,
                )

        email = UserManager.normalize_email(email)
        assert email

        username = email.split("@", 1)[0]

        with atomic():
            # Look for a user, either by ID, or email
            user = User.objects.filter(google_iap_id=user_id).first()
            if not user:
                # We explicitly don't do an OR query here, because we only want
                # to search by email if the user doesn't exist by ID. ID takes
                # precendence.
                user = User.objects.filter(email_lower=email.lower()).first()

                if user and user.google_iap_id:
                    logging.warning(
                        "Found an existing user by email (%s) who had a different "
                        "IAP user ID (%s != %s). This seems like a bug.",
                        email, user.google_iap_id, user_id
                    )

                    # We don't use this to avoid accidentally "stealing" another
                    # user
                    user = None

            if user:
                # So we previously had a user sign in by email, but not
                # via IAP, so we should set that ID
                if not user.google_iap_id:
                    user.google_iap_id = user_id
                    user.google_iap_namespace = namespace
                else:
                    # Should be caught above if this isn't the case
                    assert(user.google_iap_id == user_id)

                # Update the email as it might have changed or perhaps
                # this user was added through some other means and the
                # sensitivity of the email differs etc.
                user.email = email

                # Note we don't update the username, as that may have
                # been overridden by something post-creation
                user.save()
            else:
                with atomic():
                    # First time we've seen this user
                    user = User.objects.create(
                        google_iap_id=user_id,
                        google_iap_namespace=namespace,
                        email=email,
                        username=_generate_unused_username(username)
                    )
                    user.set_unusable_password()
                    user.save()

        return user


def _validate_iap_jwt(iap_jwt, expected_audience):
    """Validate an IAP JWT.

    Args:
      iap_jwt: The contents of the X-Goog-IAP-JWT-Assertion header.
      expected_audience: The Signed Header JWT audience. See
          https://cloud.google.com/iap/docs/signed-headers-howto
          for details on how to get this value.

    Returns:
      (user_id, user_email).
    """

    decoded_jwt = id_token.verify_token(
        iap_jwt, requests.Request(), audience=expected_audience,
        certs_url='https://www.gstatic.com/iap/verify/public_key')
    return (decoded_jwt['sub'], decoded_jwt['email'])


def _get_IAP_audience_from_settings():
    return getattr(settings, _IAP_AUDIENCE)



# Work around the fact that google.auth.default doesn't support custom
# scopes on production

# Special thanks to David Buxton for this code
# https://gist.github.com/davidwtbuxton/525924b7f06f56b8530947d55bad1c21

import datetime

from google.auth import _helpers
from google.auth import credentials
from google.auth.compute_engine import _metadata


class ServiceAccountCredentials(credentials.Scoped, credentials.Credentials):
    """Credentials for App Engine runtime using the metadata service.
    In production `google.auth.default()` returns an instance of the Compute
    Engine credentials class, which does not currently support custom oauth
    scopes, even though it uses the metadata service which does.
    """

    def __init__(self, scopes=None, default_scopes=None, service_account_id="default"):
        super().__init__()
        self._scopes = scopes
        self._default_scopes = default_scopes
        self._service_account_id = service_account_id

    def refresh(self, request):
        scopes = self._scopes if self._scopes is not None else self._default_scopes
        data = self._get_token(request, scopes)
        seconds = data["expires_in"]
        token_expiry = _helpers.utcnow() + datetime.timedelta(seconds=seconds)

        self.token = data["access_token"]
        self.expiry = token_expiry

    @classmethod
    def _get_token(cls, request, scopes=None):
        token_url = "instance/service-accounts/default/token"

        if scopes:
            if not isinstance(scopes, str):
                scopes = ",".join(scopes)

            token_url = _helpers.update_query(token_url, {"scopes": scopes})

        token_data = _metadata.get(request, token_url)

        return token_data

    @property
    def requires_scopes(self):
        return not self._scopes

    def with_scopes(self, scopes, default_scopes=None):
        return self.__class__(
            scopes=scopes, default_scopes=default_scopes, service_account_id=self._service_account_id
        )



import logging
import os
from io import (
    BytesIO,
    UnsupportedOperation,
)

import requests
from django.conf import settings
from django.core.files.storage import (
    File,
    Storage,
)
from google.cloud.exceptions import NotFound

from djangae.environment import project_id, is_production_environment

BUCKET_KEY = "CLOUD_STORAGE_BUCKET"

logger = logging.getLogger(__name__)


def _get_storage_client():
    """Gets an instance of a google CloudStorage Client

        Note: google storage python library depends on env variables read at
        module import time, so that should be set before import if overwrite needed
    """

    http = None

    if not is_production_environment():
        http = requests.Session()

    from google.cloud import storage
    return storage.Client(
        project=project_id(),
        _http=http,
    )


def _get_default_bucket_name():
    default_bucket = None
    p_id = project_id()
    if p_id:
        default_bucket = "{}.appspot.com".format(p_id)

    return default_bucket


def get_bucket_name():
    """Returns the configured bucket name

    Bucket name can be configured via settings[BUCKET_KEY]. If not set, it
    defaults to the GCP default bucket (<project_id>.appspot.com)

    Raises:
        ImproperlyConfigured: if neither configuration nor default can be retreived

    Returns:
        str -- name of the configured bucket
    """
    bucket_name = getattr(settings, BUCKET_KEY, None)
    if not bucket_name:
        bucket_name = _get_default_bucket_name()

    if not bucket_name:
        from django.core.exceptions import ImproperlyConfigured
        message = "{} not set or no default bucket configured".format(BUCKET_KEY)
        raise ImproperlyConfigured(message)

    return bucket_name


class CloudStorageFile(File):
    def __init__(self, bucket, name=None, mode="rb"):
        self._name = name
        self._mode = mode
        self._blob = bucket.blob(name)

    def read(self, num_bytes=None):
        if "r" not in self._mode:
            raise UnsupportedOperation("File open in '{}' mode is not readable".format(self._mode))
        if num_bytes:
            raise NotImplementedError("Specified argument 'num_bytes: {}' not supported".format(num_bytes))

        f = BytesIO()
        self._blob.download_to_file(f)
        return f.getvalue()

    def write(self, content):
        raise NotImplementedError("Write of CloudStorageFile object not currently supported.")


class CloudStorage(Storage):
    """
        Google Cloud Storage backend, set this as your default backend
        for ease of use, you can specify and non-default bucket in the
        constructor.

        You can modify objects access control by changing google_acl
        attribute to one of mentioned by docs (XML column):
        https://cloud.google.com/storage/docs/access-control/lists?hl=en#predefined-acl
    """
    def __init__(self, bucket_name=None, google_acl=None):
        self._bucket_name = bucket_name if bucket_name else get_bucket_name()
        self._client = None
        self._bucket = None
        self._google_acl = google_acl

    @property
    def client(self):
        """Returns the GCS client

        Returns:
            google.cloud.storage.Client -- GCS Client instance
        """
        if self._client is None:
            self._client = _get_storage_client()
        return self._client

    @property
    def bucket(self):
        if not self._bucket:
            try:
                self._bucket = self.client.get_bucket(self._bucket_name)
            except NotFound as e:
                logger.info("Bucket '{}' does not exist".format(self._bucket_name))
                raise e

        return self._bucket

    def get_valid_name(self, name):
        # App Engine doesn't properly deal with "./" and a blank upload_to argument
        # on a filefield results in ./filename so we must remove it if it's there.
        if name.startswith("./"):
            name = name.replace("./", "", 1)
        return name

    def exists(self, name):
        return bool(self.bucket.get_blob(name))

    def _save(self, name, content):
        # Not sure why, but it looks like django is not actually calling this
        name = self.get_valid_name(name)
        blob = self.bucket.blob(name)
        blob.upload_from_file(
            content.file, size=content.size, predefined_acl=self._google_acl
        )
        return name

    def _open(self, name, mode="r"):
        return CloudStorageFile(self.bucket, name=name, mode=mode)

    def size(self, name):
        blob = self.bucket.get_blob(name)
        if blob is None:
            raise NotFound("File {} does not exists".format(name))
        return blob.size

    def delete(self, name):
        """Delete an object by name

        Arguments:
            name {str} -- Name of the object to delete
        """
        return self.bucket.delete_blob(name)

    def url(self, name):
        return self.get_public_url(name)

    def get_public_url(self, name):
        """Gets the public URL of a blob

        Note: the public url is not guaranteed to be accessible. This depends on your bucket/object
        ACL and IAM. When using the gcs emulator, all objects are treated as publicly accessible

        Arguments:
            name {str} -- name of the blob

        Returns:
            str -- Public url
        """

        if is_production_environment():
            blob = self.bucket.blob(name)
            return blob.public_url
        else:
            return "{}/{}/{}".format(os.environ["STORAGE_EMULATOR_HOST"], self._bucket_name, name)



from .environment import _TASK_ENV

_TASK_NAME_HEADER = "HTTP_X_APPENGINE_TASKNAME"
_QUEUE_NAME_HEADER = "HTTP_X_APPENGINE_QUEUENAME"
_TASK_EXECUTION_COUNT_HEADER = "HTTP_X_APPENGINE_TASKEXECUTIONCOUNT"
_TASK_RETRY_COUNT_HEADER = "HTTP_X_APPENGINE_TASKRETRYCOUNT"
_APPENGINE_CRON_HEADER = "HTTP_X_APPENGINE_CRON"


def task_environment_middleware(get_response):
    def middleware(request):
        # Make sure we set the appengine headers in the environment from the
        # request.
        try:
            _TASK_ENV.task_name = request.META.get(_TASK_NAME_HEADER)
            _TASK_ENV.queue_name = request.META.get(_QUEUE_NAME_HEADER)
            _TASK_ENV.task_execution_count = request.META.get(_TASK_EXECUTION_COUNT_HEADER)
            _TASK_ENV.task_retry_count = request.META.get(_TASK_RETRY_COUNT_HEADER)
            _TASK_ENV.is_cron = bool(request.META.get(_APPENGINE_CRON_HEADER))

            return get_response(request)
        finally:
            for attr in (
                "task_name",
                "queue_name",
                "task_execution_count",
                "task_retry_count",
                "is_cron",
            ):
                setattr(_TASK_ENV, attr, None)

    return middleware


TaskEnvironmentMiddleware = task_environment_middleware



from django.db import models
from django.utils import timezone
from djangae.tasks.deferred import (
    defer_iteration_with_finalize,
    get_deferred_shard_index,
)
from djangae.test import (
    TaskFailedBehaviour,
    TestCase,
)

import time


_SHARD_COUNT = 5


class DeferIterationTestModel(models.Model):
    touched = models.BooleanField(default=False)
    finalized = models.BooleanField(default=False)
    ignored = models.BooleanField(default=False)


class DeferStringKeyModel(models.Model):
    name = models.CharField(primary_key=True, max_length=32)
    other = models.CharField(max_length=32)
    time_hit = models.DateTimeField(null=True)

    class Meta:
        ordering = (
            "-other",
        )


def callback(instance, touch=True):
    shard_index = get_deferred_shard_index()

    assert(shard_index >= 0)
    assert(shard_index < 5)

    if touch:
        instance.touched = True
    instance.save()


sporadic_error_counter = 0


def sporadic_error(instance):
    global sporadic_error_counter

    if instance.pk == 1:
        sporadic_error_counter += 1
        if sporadic_error_counter in (0, 1, 2):
            raise ValueError("Boom!")

    instance.touched = True
    instance.save()


def finalize(touch=True):
    for instance in DeferIterationTestModel.objects.all():
        instance.finalized = True
        instance.save()


def update_timestamp(instance):
    instance.time_hit = timezone.now()
    instance.save()
    time.sleep(0.1)


def noop(*args, **kwargs):
    pass


class DeferIterationTestCase(TestCase):
    def test_passing_args_and_kwargs(self):
        [DeferIterationTestModel.objects.create() for i in range(25)]

        defer_iteration_with_finalize(
            DeferIterationTestModel.objects.all(),
            callback,
            finalize,
            touch=False,  # kwarg to not touch the objects at all
            _shards=_SHARD_COUNT
        )

        self.process_task_queues()

        self.assertEqual(0, DeferIterationTestModel.objects.filter(touched=True).count())

    def test_instances_hit(self):
        [DeferIterationTestModel.objects.create() for i in range(25)]

        defer_iteration_with_finalize(
            DeferIterationTestModel.objects.all(),
            callback,
            finalize,
            _shards=_SHARD_COUNT
        )

        self.process_task_queues()

        self.assertEqual(25, DeferIterationTestModel.objects.filter(touched=True).count())
        self.assertEqual(25, DeferIterationTestModel.objects.filter(finalized=True).count())

    def test_excluded_missed(self):
        [DeferIterationTestModel.objects.create(ignored=(i < 5)) for i in range(25)]

        defer_iteration_with_finalize(
            DeferIterationTestModel.objects.filter(ignored=False),
            callback,
            finalize,
            _shards=_SHARD_COUNT
        )

        self.process_task_queues()

        self.assertEqual(5, DeferIterationTestModel.objects.filter(ignored=True).count())
        self.assertEqual(20, DeferIterationTestModel.objects.filter(touched=True).count())
        self.assertEqual(25, DeferIterationTestModel.objects.filter(finalized=True).count())

    def test_shard_continue_on_error(self):
        [DeferIterationTestModel.objects.create(pk=i + 1) for i in range(25)]

        global sporadic_error_counter
        sporadic_error_counter = 0

        defer_iteration_with_finalize(
            DeferIterationTestModel.objects.all(),
            sporadic_error,
            finalize,
            _shards=_SHARD_COUNT
        )

        self.process_task_queues(failure_behaviour=TaskFailedBehaviour.RETRY_TASK)

        self.assertEqual(25, DeferIterationTestModel.objects.filter(touched=True).count())
        self.assertEqual(25, DeferIterationTestModel.objects.filter(finalized=True).count())

    def test_shard_iterated_in_order(self):
        DeferStringKeyModel.objects.create(name="D", other="A")
        DeferStringKeyModel.objects.create(name="B", other="B")
        DeferStringKeyModel.objects.create(name="A", other="C")
        DeferStringKeyModel.objects.create(name="C", other="D")
        DeferStringKeyModel.objects.create(name="E", other="E")

        defer_iteration_with_finalize(
            DeferStringKeyModel.objects.all(),
            update_timestamp,
            noop,
            _shards=2
        )

        self.process_task_queues()

        instances = DeferStringKeyModel.objects.order_by("time_hit")

        last_id = "\0"

        for instance in instances:
            self.assertTrue(instance.pk > last_id)
            last_id = instance.pk





class LOCK_KINDS(object):
    """ The different kinds of lock which you can use.
        WEAK is not guaranteed to be robust, but can be used for situations where avoiding
        simultaneous code execution is preferable but not critical.
        STRONG is for where preventing simultaneous code execution is *required*.
    """
    WEAK = 'weak'
    STRONG = 'strong'



from django.test import (
    RequestFactory,
    override_settings,
)

from djangae.contrib import sleuth
from djangae.test import TestCase
from djangae.tasks.middleware import _TASK_NAME_HEADER
from ..views import create_datastore_backup


class DatastoreBackupViewTestCase(TestCase):
    """Tests for djangae.contrib.backup.views"""

    @override_settings(DJANGAE_BACKUP_ENABLED=False)
    def test_flag_prevents_backup(self):
        request = RequestFactory().get('/')
        request.META[_TASK_NAME_HEADER] = "test"
        with sleuth.watch('djangae.contrib.backup.views.backup_datastore') as backup_fn:
            create_datastore_backup(request)
            self.assertFalse(backup_fn.called)

    @override_settings(DJANGAE_BACKUP_ENABLED=True)
    def test_get_params_propogate(self):
        request = RequestFactory().get('/?kind=django_admin_log&bucket=foobar')
        request.META[_TASK_NAME_HEADER] = "test"
        with sleuth.fake("djangae.tasks.decorators.is_in_task", True):
            with sleuth.fake('djangae.contrib.backup.views.backup_datastore', None) as backup_fn:
                create_datastore_backup(request)
                self.assertTrue(backup_fn.called)
                self.assertEqual(
                    backup_fn.calls[0][1],
                    {
                        'bucket': 'foobar',
                        'kinds': [u'django_admin_log']
                    }
                )



from django.http import HttpResponse
from django.test import (
    Client,
    RequestFactory,
    override_settings,
)
from django.urls import (
    path,
    reverse,
)

from djangae.contrib import sleuth
from djangae.tasks.decorators import (
    csrf_exempt_if_task,
    task_only,
    task_or_superuser_only,
)
from djangae.tasks.deferred import defer
from djangae.tasks.environment import (
    is_in_cron,
    is_in_task,
    task_name,
    task_queue_name,
    task_execution_count,
    task_retry_count,
)
from djangae.tasks import environment
from djangae.tasks.middleware import _TASK_NAME_HEADER, task_environment_middleware
from djangae.test import TestCase


class TaskOnlyTestCase(TestCase):
    """ Tests for the @task_only decorator. """

    def setUp(self):
        self.factory = RequestFactory()
        super().setUp()

    def test_403_if_not_task(self):
        # If we are neither in a task or logged in as an admin, we expect a 403 response

        @task_only
        def view(request):
            return HttpResponse("Hello")

        request = self.factory.get("/")
        response = view(request)
        self.assertEqual(response.status_code, 403)

    def test_allowed_if_in_task(self):
        """ If we're in an App Engine task then it should allow the request through. """

        @task_only
        def view(request):
            return HttpResponse("Hello")

        request = self.factory.get("/")
        with sleuth.fake("djangae.tasks.decorators.is_in_task", True):
            response = view(request)

        self.assertEqual(response.status_code, 200)

    def test_allowed_if_in_cron(self):
        """ If the view is being called by the GAE cron, then it should allow the request through. """

        @task_only
        def view(request):
            return HttpResponse("Hello")

        request = self.factory.get("/")

        with sleuth.fake("djangae.tasks.decorators.is_in_cron", True):
            response = view(request)
        self.assertEqual(response.status_code, 200)


class TaskOrSuperuserOnlyTestCase(TestCase):
    """ Tests for the @task_only decorator. """

    def setUp(self):
        self.factory = RequestFactory()
        super().setUp()

    def test_403_if_not_task_or_superuser(self):
        # If we are neither in a task or logged in as an admin, we expect a 403 response

        @task_or_superuser_only
        def view(request):
            return HttpResponse("Hello")

        request = self.factory.get("/")
        response = view(request)
        self.assertEqual(response.status_code, 403)

    def test_allowed_if_in_task(self):
        """ If we're in an App Engine task then it should allow the request through. """

        @task_or_superuser_only
        def view(request):
            return HttpResponse("Hello")

        request = self.factory.get("/")
        request.META[_TASK_NAME_HEADER] = "test"

        with sleuth.fake("djangae.tasks.decorators.is_in_task", True):
            response = view(request)
        self.assertEqual(response.status_code, 200)

    def test_allowed_if_superuser(self):
        """ If we're in an App Engine task then it should allow the request through. """

        @task_or_superuser_only
        def view(request):
            return HttpResponse("Hello")

        class User(object):
            is_superuser = True
            is_authenticated = True

        request = self.factory.get("/")
        request.user = None
        response = view(request)
        self.assertEqual(response.status_code, 403)

        request.user = User()
        response = view(request)
        self.assertEqual(response.status_code, 200)


@csrf_exempt_if_task
def view(request):
    return HttpResponse("Hello")


urlpatterns = [
    path("view", view, name="test_view")
]


@override_settings(ROOT_URLCONF='djangae.tasks.tests.test_environment')
class CsrfExemptIfTaskTest(TestCase):
    def test_csrf_required_if_normal_view(self):
        """ If we're in an App Engine task then it should allow the request through. """

        client = Client(enforce_csrf_checks=True)
        response = client.post(reverse("test_view"))
        self.assertEqual(response.status_code, 403)

        response = client.post(reverse("test_view"), HTTP_X_APPENGINE_TASKNAME="test")
        self.assertEqual(response.status_code, 200)


def deferred_func():
    assert(task_name())
    assert(task_queue_name())
    assert(task_execution_count())
    assert(not is_in_cron())


class EnvironmentTests(TestCase):
    def test_task_headers_are_available_in_tests(self):
        defer(deferred_func)
        self.process_task_queues()

        # Check nothing lingers
        self.assertFalse(task_name())
        self.assertFalse(task_queue_name())
        self.assertFalse(task_execution_count())

    def test_middleware_resets_thread_locals_flags(self):
        # Set all the task environment flags that should be reset
        for attr, set_value in (
            ("task_name", "footask"),
            ("queue_name", "fooqueue"),
            ("task_execution_count", 1),
            ("task_retry_count", 1),
            ("is_cron", True),
        ):
            setattr(environment._TASK_ENV, attr, set_value)

        # All helpers should return their active value
        self.assertTrue(is_in_task())
        self.assertTrue(is_in_cron())
        self.assertEquals(task_name(), "footask")
        self.assertEquals(task_queue_name(), "fooqueue")
        self.assertEquals(task_retry_count(), 1)
        self.assertEquals(task_execution_count(), 1)

        # Running the middleware should unset flags
        def get_response(request):
            return HttpResponse()
        request = RequestFactory().get("/")
        task_environment_middleware(get_response)(request)

        # All helpers should be reset to their default
        self.assertFalse(is_in_task())
        self.assertFalse(is_in_cron())
        self.assertIsNone(task_name())
        self.assertIsNone(task_queue_name())
        self.assertIsNone(task_retry_count())
        self.assertIsNone(task_execution_count())



import logging

from djangae.test import TestCase
from djangae.contrib.security.management.commands import dumpurls
from djangae.contrib.security.middleware import (
    get_default_argument,
    find_argument_index,
    replace_default_argument
)


class DumpUrlsTests(TestCase):
    def test_dumpurls(self):
        """ Test that the `dumpurls` command runs without dying. """
        logging.debug('%s', "*" * 50)
        command = dumpurls.Command()
        command.handle()


class MiddlewareTests(TestCase):

    def test_find_argument_index(self):
        def dummy(a, b):
            pass

        self.assertEqual(find_argument_index(dummy, 'b'), 1)

    def test_get_default_argument(self):
        def dummy(a, b=2):
            pass

        self.assertEqual(get_default_argument(dummy, 'b'), 2)

    def test_replace_default_argument(self):
        def dummy(a, b=2):
            pass

        replace_default_argument(dummy, 'b', 5)
        self.assertEqual(get_default_argument(dummy, 'b'), 5)



import ast
import csv
import functools
import inspect
import json

from itertools import chain

from django.conf import settings
from django.core.management.base import BaseCommand
from django.urls import reverse, resolve

from djangae.contrib.security.commands_utils import (
    extract_views_from_urlpatterns,
    display_as_table,
    get_func_name,
    get_decorators,
    get_mixins,
    simplify_regex,
)


DEFAULT_IGNORED_MODULES = ['django', '__builtin__']


class Command(BaseCommand):
    args = "<module_to_ignore> <module_to_ignore> ..."
    help = "Displays all of the url matching routes for the project."

    def add_arguments(self, parser):
        parser.add_argument('--show_allowed_methods', action='store_true')
        parser.add_argument('--show_class_parents', action='store_true')
        parser.add_argument(
            '--output_file_type',
            nargs='?',
            choices=['json', 'csv'],
        )

    def handle(self, *args, **options):
        show_class_parents = options.get('show_class_parents')
        show_allowed_methods = options.get('show_allowed_methods')
        output_type = options.get('output_file_type')

        ignored_modules = args if args else DEFAULT_IGNORED_MODULES
        urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [''])
        view_functions = extract_views_from_urlpatterns(urlconf.urlpatterns, ignored_modules=ignored_modules)

        views = []
        for (func, regex, url_name) in view_functions:

            # Extract real function from partial
            if isinstance(func, functools.partial):
                func = func.func

            decorators_and_mixins = get_decorators(func) + get_mixins(func, ignored_modules=ignored_modules)

            view_info = dict(
                module='{0}.{1}'.format(func.__module__, get_func_name(func)),
                url=simplify_regex(regex),
                decorators=', '.join(decorators_and_mixins),
                parents='',
                allowed_http_methods='',
            )

            cbv_info = get_cbv_info(url_name)
            if cbv_info:
                view_info['parents'] = ', '.join(cbv_info['parent_class_names'])
                view_info['allowed_http_methods'] = ', '.join(cbv_info['allowed_http_methods'])
                view_info['decorators'] = ', '.join([
                    '{}: {}'.format(method_name, ', '.join(['@' + d for d in decorators]))
                    for method_name, decorators in cbv_info['decorators'].items()
                ])

            views.append(view_info)

        info = (
            "Decorators lists are not comprehensive and do not take account of other patching.\n"
        )

        headers = ['URL', 'Handler path', 'Decorators & Mixins']
        fields = ['url', 'module', 'decorators']

        if show_class_parents:
            headers += ['Parents']
            fields += ['parents']

        if show_allowed_methods:
            headers += ['Allowed Methods']
            fields += ['allowed_http_methods']

        formatted_views = []
        for view in views:
            formatted_views.append([view[field] for field in fields])

        table = display_as_table(formatted_views, headers)
        console_output = "\n{0}\n{1}".format(table, info)

        if output_type:
            console_output += _write_to_file(views, output_type)

        return console_output


def get_cbv_info(url_name):
    if url_name is None:
        return

    url = reverse(url_name)
    view = resolve(url).func
    is_class_based_view = inspect.isclass(view)

    if is_class_based_view:
        return {
            'decorators':  _get_class_decorators(view),
            'allowed_http_methods': view.http_method_names,
            'parent_class_names': [
                klass.__name__
                for klass in inspect.getmro(view)
                if klass.__name__ != 'object'
            ],
        }


def _get_class_decorators(cls):
    # from https://stackoverflow.com/a/31197273
    target = cls
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = []
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[node.name].append(name)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(inspect.getsource(target)))
    return decorators


def _write_to_file(rows, output_type='json'):
    if output_type == 'json':
        filename = 'dumpurls.json'
        with open(filename, 'w+') as f:
            f.write(json.dumps(rows, sort_keys=True, indent=2, separators=(',', ': ')))

    elif output_type == 'csv':
        filename = 'dumpurls.csv'
        with open(filename, 'w+') as f:
            csv_writer = csv.DictWriter(
                f, delimiter=',', fieldnames={field for field in chain(*[r.keys() for r in rows])},
                dialect='excel', quotechar='"'
            )
            csv_writer.writeheader()
            csv_writer.writerows(rows)

    return '\nUrls dumped to: {}'.format(filename)



import os
from setuptools import setup, find_packages


NAME = 'djangae'
PACKAGES = find_packages()
DESCRIPTION = 'Django integration with Google App Engine'
URL = "https://github.com/potatolondon/djangae"
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
AUTHOR = 'Potato London Ltd.'

EXTRAS = {
    "test": ["beautifulsoup4"],
}

setup(
    name=NAME,
    version='2.0.0rc1',
    packages=PACKAGES,

    # metadata for upload to PyPI
    author=AUTHOR,
    author_email='mail@p.ota.to',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=["django", "Google App Engine", "GAE"],
    url=URL,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],

    include_package_data=True,
    # dependencies
    install_requires=[
        'django>=2.2,<=3.3.*',
        'django-gcloud-connectors>=0.3.2',
        'google-api-python-client>=1.7.11',
        'google-cloud-tasks>=1.5.0,<2.0.0',
        'psutil>=5.7.3',
        # requests required by cloud storage file backend
        'requests>=2.22.0',
        # required for iap backend
        'cryptography==3.4.6',
        'python-jose[cryptography]==3.2.0',
    ],
    extras_require=EXTRAS,
    tests_require=EXTRAS['test'],
)




from djangae.contrib.googleauth import (
    _GOOG_AUTHENTICATED_USER_EMAIL_HEADER,
    _GOOG_AUTHENTICATED_USER_ID_HEADER,
    _GOOG_JWT_ASSERTION_HEADER
)
import hashlib
import logging
import os

from django import forms
from django.conf import settings
from django.contrib.auth import (
    BACKEND_SESSION_KEY,
    HASH_SESSION_KEY,
    REDIRECT_FIELD_NAME,
    _get_user_session_key,
    get_backends,
    constant_time_compare,
    load_backend,
    login,
    logout,
)
from django.contrib.auth.middleware import AuthenticationMiddleware
from django.http import HttpResponseRedirect
from django.shortcuts import render, redirect
from django.urls.base import reverse
from django.utils.functional import SimpleLazyObject
from django.utils.http import urlencode


from djangae.environment import is_production_environment

from .backends.iap import IAPBackend
from .backends.oauth2 import OAuthBackend
from .models import OAuthUserSession


_OAUTH_LINK_EXPIRY_SETTING = "GOOGLEAUTH_LINK_OAUTH_SESSION_EXPIRY"


def get_user_object(request):
    """
    Return the user model instance associated with the given request session.
    If no user is retrieved, return an instance of `AnonymousUser`.
    """
    from .models import AnonymousUser

    user = None
    try:
        user_id = _get_user_session_key(request)
        backend_path = request.session[BACKEND_SESSION_KEY]
    except KeyError:
        pass
    else:
        if backend_path in settings.AUTHENTICATION_BACKENDS:
            backend = load_backend(backend_path)
            user = backend.get_user(user_id)
            # Verify the session
            if hasattr(user, 'get_session_auth_hash'):
                session_hash = request.session.get(HASH_SESSION_KEY)
                session_hash_verified = session_hash and constant_time_compare(
                    session_hash,
                    user.get_session_auth_hash()
                )
                if not session_hash_verified:
                    request.session.flush()
                    user = None

    return user or AnonymousUser()


def get_user(request):
    if not hasattr(request, '_cached_user'):
        request._cached_user = get_user_object(request)
    return request._cached_user


class AuthenticationMiddleware(AuthenticationMiddleware):
    def process_request(self, request):
        assert hasattr(request, 'session'), (
            "The djangae.contrib.googleauth middleware requires session middleware "
            "to be installed. Edit your MIDDLEWARE%s setting to insert "
            "'django.contrib.sessions.middleware.SessionMiddleware' before "
            "'djangae.contrib.googleauth.middleware.AuthenticationMiddleware'."
        ) % ("_CLASSES" if settings.MIDDLEWARE is None else "")

        request.user = SimpleLazyObject(lambda: get_user(request))

        # See if the handling view is marked with the auth_middleware_exempt
        # decorator, and return if so.
        if request.resolver_match:
            func = request.resolver_match.func
            exempt = getattr(func, "_auth_middleware_exempt", False)
            if exempt:
                return None

        backend_str = request.session.get(BACKEND_SESSION_KEY)

        if request.user.is_authenticated:
            if backend_str and isinstance(load_backend(backend_str), OAuthBackend):

                # Should we link the Django session to the OAuth session? In most cases we shouldn't
                # as oauth would've been used for identification at login only.
                expire_session = getattr(settings, _OAUTH_LINK_EXPIRY_SETTING, False)

                if expire_session:
                    # The user is authenticated with Django, and they use the OAuth backend, so they
                    # should have a valid oauth session
                    oauth_session = OAuthUserSession.objects.filter(
                        pk=request.user.google_oauth_id
                    ).first()

                    # Their oauth session does not exist, so let's log them out
                    if not oauth_session:
                        logout(request)
                        return None

                    # Their oauth session expired but we still have an active user session
                    if not oauth_session.is_valid:
                        return redirect(
                            reverse("googleauth_oauth2login") + '?' + urlencode(dict(next=request.path))
                        )

            elif backend_str and isinstance(load_backend(backend_str), IAPBackend):
                if not IAPBackend.can_authenticate(request):
                    logout(request)
        else:
            backends = get_backends()
            try:
                iap_backend = next(filter(lambda be: isinstance(be, IAPBackend), backends))
            except StopIteration:
                iap_backend = None

            # Try to authenticate with IAP if the headers
            # are available
            if iap_backend and IAPBackend.can_authenticate(request):
                user = iap_backend.authenticate(request)
                if user and user.is_authenticated:
                    user.backend = 'djangae.contrib.googleauth.backends.iap.%s' % IAPBackend.__name__
                    login(request, user)


class ProfileForm(forms.Form):
    email = forms.EmailField()


_CREDENTIALS_FILE = os.path.join(
    settings.BASE_DIR, ".iap-credentials"
)


def _login_view(request):
    if request.method == "POST":
        form = ProfileForm(request.POST)
        if form.is_valid():
            # We write a credentials file for 2 reasons:
            # 1. It will persist across local server restarts.
            # 2. It will blow up on production, as the local folder
            #    is not writable.
            with open(_CREDENTIALS_FILE, "w") as f:
                f.write(
                    "%s\n" % (
                        form.cleaned_data["email"]
                    )
                )

            dest = request.GET.get(REDIRECT_FIELD_NAME, "/")
            return HttpResponseRedirect(dest)
    else:
        form = ProfileForm()

    subs = {
        "form": form
    }

    return render(request, "googleauth/dev_login.html", subs)


def id_from_email(email):
    """
        Just generates a predictable user ID from the email entered
    """
    md5 = hashlib.md5()
    md5.update(email.encode("utf-8"))

    # Truncate to 32-bit
    return int(md5.hexdigest(), 16) & 0xFFFFFFFF


def local_iap_login_middleware(get_response):
    def middleware(request):
        request._through_local_iap_middleware = True

        if is_production_environment():
            logging.warning(
                "local_iap_login_middleware is for local development only, "
                "and will not work on production. "
                "You should remove it from your MIDDLEWARE setting"
            )
            response = get_response(request)
        elif request.path == "/_dj/login/":
            response = _login_view(request)
        elif request.path == "/_dj/logout/":
            if os.path.exists(_CREDENTIALS_FILE):
                os.remove(_CREDENTIALS_FILE)

            if REDIRECT_FIELD_NAME in request.GET:
                return HttpResponseRedirect(request.GET[REDIRECT_FIELD_NAME])
            else:
                return HttpResponseRedirect("/_dj/login/")
        else:
            if os.path.exists(_CREDENTIALS_FILE):
                # Update the request headers with the stored credentials
                with open(_CREDENTIALS_FILE, "r") as f:
                    data = f.read()
                    email = data.strip()

                    request.META[_GOOG_JWT_ASSERTION_HEADER] = "JWT TOKEN"
                    request.META[_GOOG_AUTHENTICATED_USER_ID_HEADER] = "auth.example.com:%s" % id_from_email(email)
                    request.META[_GOOG_AUTHENTICATED_USER_EMAIL_HEADER] = "auth.example.com:%s" % email

            response = get_response(request)

        return response
    return middleware


LocalIAPLoginMiddleware = local_iap_login_middleware




from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from ..models import (
    AbstractGoogleUser,
    Group,
    UserManager,
    UserPermission,
)
from . import _find_atomic_decorator, _generate_unused_username
from .base import BaseBackend

User = get_user_model()


class OAuthBackend(BaseBackend):
    def authenticate(self, request, **kwargs):
        atomic = _find_atomic_decorator(User)

        if not issubclass(User, AbstractGoogleUser):
            raise ImproperlyConfigured(
                "OAuthBackend requires AUTH_USER_MODEL to be a "
                " subclass of djangae.contrib.auth.models.AbstractGoogleUser."
            )

        oauth_session = kwargs.get("oauth_session")

        if (not oauth_session) or (not oauth_session.is_valid):
            return

        # FIXME: Refresh the token if it's close to expiry?

        profile = oauth_session.profile

        email = UserManager.normalize_email(profile["email"])
        assert email
        username = email.split("@", 1)[0]

        with atomic():
            # Look for a user, either by oauth session ID, or email
            user = User.objects.filter(
                google_oauth_id=oauth_session.pk
            ).first()

            if not user:
                # Only fallback to email if we didn't find by session ID
                user = User.objects.filter(email_lower=email.lower()).first()

            # So we previously had a user sign in by email, but not
            # via OAuth, so let's update their user with their oauth
            # session ID
            if user:
                if not user.google_oauth_id:
                    user.google_oauth_id = oauth_session.pk
                else:
                    assert(user.google_oauth_id == oauth_session.pk)
                    # We got the user by google_oauth_id, but their email
                    # might have changed (maybe), so update that just in case
                    user.email = email
                if not user.username:
                    user.username = _generate_unused_username(username)
                user.save()
            else:
                # First time we've seen this user
                user = User.objects.create(
                    google_oauth_id=oauth_session.pk,
                    email=email,
                    username=_generate_unused_username(username)
                )
                user.set_unusable_password()
                user.save()

        return user

    def user_can_authenticate(self, user):
        """
        Reject users with is_active=False. Custom user models that don't have
        that attribute are allowed.
        """
        is_active = getattr(user, 'is_active', None)
        return is_active or is_active is None

    def get_user(self, user_id):
        try:
            user = User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None
        return user if self.user_can_authenticate(user) else None

    def get_user_permissions(self, user_obj, obj=None):
        qs = UserPermission.objects.filter(
            user_id=user_obj.pk
        ).values_list("permission", flat=True)

        if obj:
            qs = qs.filter(obj_id=obj.pk)

        return list(qs)

    def get_group_permissions(self, user_obj, obj=None):
        perms = set()
        qs = Group.objects.filter(users__contains=user_obj)
        for group in qs:
            perms.update(group.permissions)

        return list(perms)



from django.utils.module_loading import autodiscover_modules

from . import model_document
from .index import Index  # noqa
from .document import Document  # noqa
from .model_document import (  # noqa
    ModelDocument,
    register,
)

from .fields import (  # noqa
    IntegrityError,
    TextField,
    NumberField,
    DateField,
    FuzzyTextField
)

default_app_config = 'djangae.contrib.search.apps.SearchConfig'


def autodiscover():
    # This will find all the search.py modules and add them to model_document._registry
    # The act of importing them will call any register(model, document) statements
    # in that file.
    autodiscover_modules('search', register_to=model_document)



# STANDARD LIB
from functools import wraps

# LIBRARIES
from django.conf import settings
from django.test.utils import override_settings


def without_security_middleware(func):
    """ Decorator for disabling djangae.contrib.security.middleware.AppEngineSecurityMiddleware.
    """
    # This is sometimes useful in tests because if you patch something which is patched by the
    # middleware then usually the test patch is applied first, THEN the middleware runs (usually
    # as the result of a self.client.get/post call), thus applying a decorator to the patched
    # function.  The decorator's nested function stores a reference to the original (patched)
    # function, and so when the patch in the test is undone it is actually still being used because
    # the decorator in the middleware is still referencing it.  This decorator is for
    # use in tests where you are patching one of the things which is patched in the middleware,
    # thus allowing you to avoid the afore-described issue.
    middleware = list(settings.MIDDLEWARE_CLASSES[:])
    middleware.remove('djangae.contrib.security.middleware.AppEngineSecurityMiddleware')

    @wraps(func)
    def _wrapped(*args, **kwargs):
        with override_settings(MIDDLEWARE_CLASSES=middleware):
            return func(*args, **kwargs)

    return _wrapped



from django.db import models

from djangae import patches  # noqa


class DeferIterationMarker(models.Model):
    """
        Marker to keep track of sharded defer
        iteration tasks
    """

    # Set to True when all shards have been deferred
    is_ready = models.BooleanField(default=False)

    shard_count = models.PositiveIntegerField(default=0)
    shards_complete = models.PositiveIntegerField(default=0)

    delete_on_completion = models.BooleanField(default=True)

    created = models.DateTimeField(auto_now_add=True)
    callback_name = models.CharField(max_length=100)
    finalize_name = models.CharField(max_length=100)

    class Meta:
        app_label = "djangae"

    @property
    def is_finished(self):
        return self.is_ready and self.shard_count == self.shards_complete

    def __unicode__(self):
        return "Background Task (%s -> %s) at %s" % (
            self.callback_name,
            self.finalize_name,
            self.created
        )



import unittest

from djangae.contrib import sleuth
from djangae.contrib.backup.tasks import (
    AUTH_SCOPES,
    _get_valid_export_kinds,
    backup_datastore,
)
from djangae.test import TestCase
from django.db import models
from django.test import override_settings
from google.auth import app_engine


class MockUser(models.Model):
    class Meta:
        app_label = "backup"


class MockLogEntry(models.Model):
    class Meta:
        app_label = "backup"


def mock_get_app_models(**kwargs):
    return [
        MockLogEntry,
        MockUser,
    ]


class GetValidExportModelsTestCase(TestCase):
    """Tests focused on djangae.contrib.backup.tasks._get_valid_export_models"""

    @override_settings(DJANGAE_BACKUP_EXCLUDE_MODELS=['backup_logentry'])
    @sleuth.switch('django.apps.apps.get_models', mock_get_app_models)
    def test_models_filtered(self):
        valid_models = _get_valid_export_kinds(
            ['backup_logentry', 'backup_mockuser']
        )
        self.assertNotIn('backup_logentry', valid_models)
        self.assertIn('backup_mockuser', valid_models)

    @override_settings(DJANGAE_BACKUP_EXCLUDE_APPS=['django'])
    @sleuth.switch('django.apps.apps.get_models', mock_get_app_models)
    def test_apps_filtered(self):
        valid_models = _get_valid_export_kinds(
            ['backup_logentry', 'backup_mockuser']
        )
        self.assertIn('backup_mockuser', valid_models)
        self.assertNotIn('django_admin_log', valid_models)

    def test_models_using_the_same_table_only_listed_once(self):
        class ModelFoo(models.Model):
            class Meta:
                db_table = "foo"

        class ModelBar(models.Model):
            class Meta:
                db_table = "foo"

        def mock_get_app_models(**kwargs):
            return [ModelFoo, ModelBar]

        with sleuth.switch('django.apps.apps.get_models', mock_get_app_models):
            valid_models = _get_valid_export_kinds()
            self.assertEquals(['foo'], valid_models)

    @override_settings(DJANGAE_BACKUP_EXCLUDE_MODELS=['backup_logentry'])
    @sleuth.switch('django.apps.apps.get_models', mock_get_app_models)
    def test_kinds_are_deduplicated(self):
        valid_models = _get_valid_export_kinds(kinds=[
            'backup_logentry',
            'backup_logentry',
            'backup_mockuser',
            'backup_mockuser',
        ])
        self.assertEquals(['backup_mockuser'], valid_models)


class BackupTestCase(TestCase):

    @override_settings(DJANGAE_BACKUP_ENABLED=True)
    @unittest.skip("Skipped until we come up with a way to test authentication locally")
    def test_ok(self):
        """Lightweight end-to-end flow test of backup_datastore."""
        with sleuth.switch(
            'djangae.contrib.backup.tasks._get_authentication_credentials',
            lambda: app_engine.Credentials(scopes=AUTH_SCOPES)
        ):
            with sleuth.switch(
                'googleapiclient.http.HttpRequest.execute', lambda x: True
            ) as mock_fn:
                kinds = ['backup_mockuser']
                backup_datastore(kinds=kinds)
                self.assertTrue(mock_fn.called)



from gcloudc.db.backends.datastore.expressions import Scatter


def _find_random_keys(queryset, shard_count):
    OVERSAMPLING_FACTOR = 32

    return list(
        queryset.model.objects.order_by(Scatter()).values_list("pk", flat=True)[
            :(shard_count * OVERSAMPLING_FACTOR)
        ]
    )


def find_key_ranges_for_queryset(queryset, shard_count):
    """
        Given a queryset and a number of shard. This function makes use
        of the __scatter__ property to return a list of key ranges
        for sharded iteration.
    """

    if shard_count > 1:
        # Use the scatter property to generate shard points
        random_keys = _find_random_keys(queryset, shard_count)

        if not random_keys:
            # No random keys? Don't shard
            key_ranges = [(None, None)]
        else:
            random_keys.sort()

            # We have enough random keys to shard things
            if len(random_keys) >= shard_count:
                index_stride = len(random_keys) / float(shard_count)
                split_keys = [random_keys[int(round(index_stride * i))] for i in range(1, shard_count)]
            else:
                split_keys = random_keys

            key_ranges = [(None, split_keys[0])] + [
                (split_keys[i], split_keys[i + 1]) for i in range(len(split_keys) - 1)
            ] + [(split_keys[-1], None)]
    else:
        # Don't shard
        key_ranges = [(None, None)]

    return key_ranges



import os
from typing import Optional

from djangae.utils import memoized

# No SDK imports allowed in module namespace because `./manage.py runserver`
# imports this before the SDK is added to sys.path. See bugs #899, #1055.


def application_id() -> str:
    # Fallback to example on local or if this is not specified in the
    # environment already
    result = os.environ.get("GAE_APPLICATION", "e~example").split("~", 1)[-1]
    return result


def is_production_environment() -> bool:
    return not is_development_environment()


def is_development_environment() -> bool:
    return 'GAE_ENV' not in os.environ or os.environ['GAE_ENV'] != 'standard'


def gae_version() -> Optional[str]:
    """Returns the current GAE version."""
    return os.environ.get('GAE_VERSION')


@memoized
def get_application_root() -> str:
    """Traverse the filesystem upwards and return the directory containing app.yaml"""
    from django.conf import settings  # Avoid circular

    path = os.path.dirname(os.path.abspath(__file__))
    app_yaml_path = os.environ.get('DJANGAE_APP_YAML_LOCATION', None)

    # If the DJANGAE_APP_YAML_LOCATION variable is setup, will try to locate
    # it from there.
    if (app_yaml_path is not None and
            os.path.exists(os.path.join(app_yaml_path, "app.yaml"))):
        return app_yaml_path

    # Failing that, iterates over the parent folders until it finds it,
    # failing when it gets to the root
    while True:
        if os.path.exists(os.path.join(path, "app.yaml")):
            return path
        else:
            parent = os.path.dirname(path)
            if parent == path:  # Filesystem root
                break
            else:
                path = parent

    # Use the Django base directory as a fallback. We search for app.yaml
    # first because that will be the "true" root of the GAE app
    return settings.BASE_DIR


def default_gcs_bucket_name() -> str:
    return "%s.appspot.com" % application_id()


def default_app_host() -> str:
    """Returns the default HOST for the application.
    Fallbacks to example.appspost.com on local
    """

    return "%s.appspot.com" % application_id()


def app_host() -> str:
    """Returns the default HOST for the application.
    Fallbacks to example-dot-example.appspost.com on local
    """
    version = gae_version() or 'example'
    return "{}-dot-{}".format(version, default_app_host)


def project_id() -> str:
    # Environment variable will exist on production servers
    # fallback to "example" locally if it doesn't exist
    return os.environ.get("GOOGLE_CLOUD_PROJECT", "example")



import datetime

from django.conf import settings
from djangae import environment

SETTINGS_PREFIX = "DJANGAE_BACKUP_"


def get_backup_setting(name, required=True, default=None):
    settings_name = "{}{}".format(SETTINGS_PREFIX, name)
    if required and not hasattr(settings, settings_name):
        raise Exception("{} is required".format(settings_name))

    return getattr(settings, settings_name, default)


def get_gcs_bucket():
    """Get a bucket from DJANGAE_BACKUP_GCS_BUCKET setting. Defaults to the
    default application bucket with 'djangae-backups' appended.

    Raises an exception if DJANGAE_BACKUP_GCS_BUCKET is missing and there is
    no default bucket.
    """
    # we don't use `get_backup_setting` here as default would be a callable
    # redundantly called if the setting was specified.
    try:
        bucket = settings.DJANGAE_BACKUP_GCS_BUCKET
    except AttributeError:
        bucket = environment.default_gcs_bucket_name()

        if bucket:
            bucket = '{}/djangae-backups'.format(bucket)

    if not bucket:
        raise Exception('No DJANGAE_BACKUP_GCS_BUCKET or default bucket')

    return bucket


def get_backup_path(bucket=None):
    """
    Returns the full path to write the backup to in GCS. This looks
    something like
    `gs://example.appspot.com/scheduled-backups/2018-20-10/202059`

    Bucket can be provided as a kwarg (e.g. passed in from a GET param),
    otherwise we fallback to the `DJANGAE_BACKUP_GCS_BUCKET` setting.
    """
    bucket = bucket if bucket else get_gcs_bucket()
    dt = datetime.datetime.utcnow()
    return 'gs://{}/{:%Y%m%d}/{:%H%M%S}'.format(bucket, dt, dt)



import logging
from importlib import import_module

from django.conf import settings
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from djangae import environment
from djangae.tasks import decorators
from djangae.core.signals import module_started, module_stopped


logger = logging.getLogger(__name__)


def warmup(request):
    """
        Provides default procedure for handling warmup requests on App
        Engine. Just add this view to your main urls.py.
    """
    for app in settings.INSTALLED_APPS:
        for name in ('urls', 'views', 'models'):
            try:
                import_module('%s.%s' % (app, name))
            except ImportError:
                pass
    return HttpResponse("OK")


def start(request):
    module_started.send(sender=__name__, request=request)
    return HttpResponse("OK")


def stop(request):
    module_stopped.send(sender=__name__, request=request)
    return HttpResponse("OK")


@csrf_exempt
def deferred(request):
    from google.appengine.ext.deferred.deferred import (
        run,
        SingularTaskFailure,
        PermanentTaskFailure
    )

    response = HttpResponse()

    if 'HTTP_X_APPENGINE_TASKEXECUTIONCOUNT' in request.META:
        logger.debug("[DEFERRED] Retry %s of deferred task", request.META['HTTP_X_APPENGINE_TASKEXECUTIONCOUNT'])

    if 'HTTP_X_APPENGINE_TASKNAME' not in request.META:
        logger.critical('Detected an attempted XSRF attack. The header "X-AppEngine-Taskname" was not set.')
        response.status_code = 403
        return response

    in_prod = environment.is_production_environment()

    if in_prod and request.META.get("REMOTE_ADDR") != "0.1.0.2":
        logger.critical('Detected an attempted XSRF attack. This request did not originate from Task Queue.')
        response.status_code = 403
        return response

    try:
        run(request.body)
    except SingularTaskFailure:
        logger.debug("Failure executing task, task retry forced")
        response.status_code = 408
    except PermanentTaskFailure:
        logger.exception("Permanent failure attempting to execute task")

    return response


@decorators.task_only
def clearsessions(request):
    engine = import_module(settings.SESSION_ENGINE)
    try:
        engine.SessionStore.clear_expired()
    except NotImplementedError:
        logger.exception(
            "Session engine '%s' doesn't support clearing "
            "expired sessions.\n", settings.SESSION_ENGINE
        )
    return HttpResponse("OK")



from djangae import checks
from djangae.test import TestCase
from django.test.utils import override_settings


class ChecksTestCase(TestCase):

    def test_csrf_check(self):
        errors = checks.check_session_csrf_enabled()
        self.assertEqual(len(errors), 0)

        with override_settings(CSRF_USE_SESSIONS=False):
            errors = checks.check_session_csrf_enabled()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].id, 'djangae.E001')

    def test_csp_report_check(self):
        errors = checks.check_csp_is_not_report_only()
        self.assertEqual(len(errors), 0)

        with override_settings(CSP_REPORT_ONLY=True):
            errors = checks.check_csp_is_not_report_only()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].id, 'djangae.E002')

    def test_csp_unsafe_check(self):
        errors = checks.check_csp_sources_not_unsafe()
        self.assertEqual(len(errors), 0)

        csp_settings = {k: ["'unsafe-inline'"] for k in checks.CSP_SOURCE_NAMES}
        with override_settings(**csp_settings):
            errors = checks.check_csp_sources_not_unsafe()
            self.assertEqual(len(errors), 9)

            # this assumes that errors come through in the same order as
            # checks.CSP_SOURCE_NAMES
            for idx, err in enumerate(errors):
                self.assertEqual(err.id, 'djangae.E1%02d' % idx)

    def test_template_loader_present(self):
        template_setting = [{
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'OPTIONS': {'loaders': [
                ('django.template.loaders.cached.Loader', ('django.template.loaders.filesystem.Loader',)),
            ]},
        }]
        with override_settings(TEMPLATES=template_setting):
            errors = checks.check_cached_template_loader_used()
            self.assertEqual(len(errors), 0)

    def test_template_loader_missing(self):
        template_setting = [{
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'OPTIONS': {'loaders': [('django.template.loaders.filesystem.Loader',)]},
        }]
        with override_settings(TEMPLATES=template_setting):
            errors = checks.check_cached_template_loader_used()
            self.assertEqual(len(errors), 1)
            self.assertEqual(errors[0].id, 'djangae.E003')

    def test_template_loader_skips_non_django_backends(self):
        template_setting = [{
            'BACKEND': 'nondjango.templates',
            'OPTIONS': {'loaders': [('nondjango.templates.Loader',)]},
        }]
        with override_settings(TEMPLATES=template_setting):
            errors = checks.check_cached_template_loader_used()
            self.assertEqual(len(errors), 0)
